#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains custom Qt tree widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Signal, QRect, QSize, QModelIndex
from Qt.QtWidgets import QApplication, QSizePolicy, QTreeWidget, QTreeWidgetItem, QAbstractItemView, QStyleOption
from Qt.QtWidgets import QWhatsThis
from Qt.QtGui import QColor, QPalette, QPen, QBrush, QPainter

from tpDcc import dcc
from tpDcc.managers import resources
from tpDcc.libs.python import path, fileio, folder
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import layouts, buttons, search, lineedit


class TreeWidget(QTreeWidget, object):

    ITEM_WIDGET = QTreeWidgetItem
    ITEM_WIDGET_SIZE = None

    def __init__(self, parent=None):
        super(TreeWidget, self).__init__(parent)

        self._auto_add_sub_items = True
        self._title_text_index = 0
        self._text_edit = True
        self._edit_state = None
        self._current_name = None
        self._old_name = None
        self._current_item = None
        self._last_item = None
        self._drop_indicator_rect = QRect()
        self._drop_indicator_position = None
        self._name_filter = None

        self.setIndentation(25)
        self.setExpandsOnDoubleClick(False)
        self.setSortingEnabled(True)
        self.sortByColumn(0, Qt.AscendingOrder)

        if dcc.is_maya():
            self.setAlternatingRowColors(dcc.get_version() < 2016)
        if not dcc.is_maya() and not not dcc.is_nuke():
            palette = QPalette()
            palette.setColor(palette.Highlight, Qt.gray)
            self.setPalette(palette)

        self.itemActivated.connect(self._on_item_activated)
        self.itemChanged.connect(self._on_item_changed)
        self.itemSelectionChanged.connect(self._on_item_selection_changed)
        self.itemClicked.connect(self._on_item_clicked)
        self.itemExpanded.connect(self._on_item_expanded)
        self.itemCollapsed.connect(self._on_item_collapsed)

    # ============================================================================================================
    # PROPERTIES
    # ============================================================================================================

    @property
    def current_item(self):
        return self._current_item

    @property
    def current_name(self):
        return self._current_name

    @property
    def edit_state(self):
        return self._edit_state

    @edit_state.setter
    def edit_state(self, flag):
        self._edit_state = flag

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def paintEvent(self, event):
        painter = QPainter(self.viewport())
        self.drawTree(painter, event.region())
        self._paint_drop_indicator(painter)

    def mousePressEvent(self, event):

        modifiers = QApplication.keyboardModifiers()
        if modifiers == Qt.AltModifier:
            pos = self.mapToGlobal((self.rect().topLeft()))
            QWhatsThis.showText(pos, self.whatsThis())
            return

        super(TreeWidget, self).mousePressEvent(event)

        item = self.itemAt(event.pos())
        if not item:
            self._clear_selection()
        else:
            self._current_item = item

    def mouseDoubleClickEvent(self, event):
        position = event.pos()
        index = self.indexAt(position)
        self.doubleClicked.emit(index)

    def dragMoveEvent(self, event):
        item = self.itemAt(event.pos())

        if item:
            index = self.indexFromItem(item)
            rect = self.visualRect(index)
            rect_left = self.visualRect(index.sibling(index.row(), 0))
            rect_right = self.visualRect(index.sibling(index.row(), self.header().logicalIndex(self.columnCount() - 1)))
            self._drop_indicator_position = self._position(event.pos(), rect)
            if self._drop_indicator_position == self.AboveItem:
                self._drop_indicator_rect = QRect(
                    rect_left.left(), rect_left.top(), rect_right.right() - rect_left.left(), 0)
                event.accept()
            elif self._drop_indicator_position == self.BelowItem:
                self._drop_indicator_rect = QRect(
                    rect_left.left(), rect_left.bottom(), rect_right.right() - rect_left.left(), 0)
                event.accept()
            elif self._drop_indicator_position == self.OnItem:
                self._drop_indicator_rect = QRect(
                    rect_left.left(), rect_left.top(), rect_right.right() - rect_left.left(), rect.height())
                event.accept()
            else:
                self._drop_indicator_rect = QRect()

            self.model().setData(index, self._drop_indicator_position, Qt.UserRole)

        self.viewport().update()

        super(TreeWidget, self).dragMoveEvent(event)

    def addTopLevelItem(self, item):
        super(TreeWidget, self).addTopLevelItem(item)

        if hasattr(item, 'widget'):
            if hasattr(item, 'column'):
                self.setItemWidget(item, item.column, item.widget)
            else:
                self.setItemWidget(item, 0, item.widget)

    def insertTopLevelItem(self, index, item):
        super(TreeWidget, self).insertTopLevelItem(index, item)

        if hasattr(item, 'widget'):
            if hasattr(item, 'column'):
                self.setItemWidget(item, item.column, item.widget)
            else:
                self.setItemWidget(item, 0, item.widget)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def item_at(self, pos):
        """
        Returns a pointer to the item at the coordinates p
        The coordinates are relative to the tree widget's viewport
        :param pos: QPoint
        :return: QTreeWidgetItem
        """

        index = self.indexAt(pos)
        return self.item_from_index(index)

    def item_from_index(self, index):
        """
        Return a pointer to the LibraryItem associated with the given model index
        :param index: QModelIndex
        :return: QTreeWidgetItem
        """

        return self.itemFromIndex(index)

    def unhide_items(self):
        """
        Unhide all tree items
        """

        for i in range(self.topLevelItemCount()):
            item = self.topLevelItem(i)
            self.setItemHidden(item, False)

    def filter_names(self, filter_text):
        """
        Hides all tree items with the given text
        :param filter_text: str, text used to filter tree items
        """

        self._name_filter = filter_text.strip(' ')

        self.unhide_items()

        for i in range(self.topLevelItemCount()):
            item = self.topLevelItem(i)
            text = str(item.text(self._title_text_index))
            filter_text = str(filter_text).strip(' ')

            # If the filter text is not found on the item text, we hide the item
            if text.find(filter_text) == -1:
                self.setItemHidden(item, True)

    def get_tree_item_name(self, tree_item):
        """
        Returns a list with all the column names of the given QTreeWidgetItem
        :param tree_item: QTreeWidgetItem, item we want to retrive name of
        :return: list<str>
        """

        try:
            # When selecting an item in the tree and refreshing a C++ wrapped will raise
            count = QTreeWidgetItem.columnCount(tree_item)
        except Exception:
            count = 0

        name = list()
        for i in range(count):
            name.append(str(tree_item.text(i)))

        return name

    def get_tree_item_path_string(self, tree_item):
        """
        Returns full path of the given QTreeWidgetItem from its parent into it with the following format:
        "parent/parent/item"
        :param tree_item:
        :return:
        """

        parents = self.get_tree_item_path(tree_item)
        parent_names = self.get_tree_item_names(parents)

        if not parent_names:
            return
        if len(parent_names) == 1 and not parent_names[0]:
            return

        names = list()

        for name in parent_names:
            names.append(name[0])

        names.reverse()

        item_path_str = '/'.join(names)

        return item_path_str

    def get_tree_item_path(self, tree_item):
        """
        Return the tree path of the given QTreeWidgetItem in a list starting from the given item
        :param tree_item: QTreeWidgetItem
        :return: list<QTreeWidgetItem>
        """

        if not tree_item:
            return

        parent_items = list()
        parent_items.append(tree_item)

        try:
            # When selecting an item in the tree and refreshing a C++ wrapped will raise
            parent_item = tree_item.parent()
        except Exception:
            parent_item = None

        while parent_item:
            parent_items.append(parent_item)
            parent_item = parent_item.parent()

        return parent_items

    def get_tree_item_names(self, tree_items):
        """
        Returns a list with the names of the given QTreeWidgetItems
        :param tree_items: list<QTreeWidgetItem>
        :return: list<str>
        """

        item_names = list()

        if not tree_items:
            return item_names

        for tree_item in tree_items:
            name = self.get_tree_item_name(tree_item)
            if name:
                item_names.append(name)

        return item_names

    def get_tree_item_children(self, tree_item):
        """
        Returns all child items of the given QTreeWidgetItem
        :param tree_item: QTreeWidgetItem
        :return: list<QTreeWidgetItem>
        """

        count = tree_item.childCount()
        items = list()
        for i in range(count):
            items.append(tree_item.child(i))

        return items

    def delete_empty_children(self, tree_item):
        """
        Deletes all given QTreeWidget child items that are empty (has no text)
        :param tree_item: QTreeWidgetItem
        """

        count = tree_item.childCount()
        if count <= 0:
            return

        for i in range(count):
            item = tree_item.child(i)
            if item and not item.text(0):
                item = tree_item.takeChild(i)
                del item

    def delete_tree_item_children(self, tree_item):
        """
        Deletes all given QTreeWidget chlid items
        :param tree_item: QTreeWidgetItem
        """

        count = tree_item.childCount()
        if count <= 0:
            return

        children = tree_item.takeChildren()
        for child in children:
            del child

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _position(self, pos, rect):
        """
        Internal function that returns whether the cursor is over, below or on an item
        :param pos: QPos
        :param rect: QRect
        :return: QAbstractItemView.DropIndicatorPosition
        """

        r = QAbstractItemView.OnViewport

        # NOTE: margin * 2 MUST be smaller than row height, otherwise drop OnItem rect will not show
        margin = 5

        if pos.y() - rect.top() < margin:
            r = QAbstractItemView.AboveItem
        elif rect.bottom() - pos.y() < margin:
            r = QAbstractItemView.BelowItem
        elif pos.y() - rect.top() > margin and rect.bottom() - pos.y() > margin:
            r = QAbstractItemView.OnItem

        return r

    def _drop_on(self, event_list):
        """
        Internal function that checks whether or not event list contains a valid drop operation
        :param event_list: list
        :return: bool
        """

        event, row, col, index = event_list
        root = self.rootIndex()

        if self.viewport().rect().contains(event.pos()):
            index = self.indexAt(event.pos())
            if not index.isValid() or not self.visualRect(index).contains(event.pos()):
                index = root

        if index != root:
            if self._drop_indicator_position == self.AboveItem:
                # Drop Above item
                row = index.row()
                col = index.column()
                index = index.parent()
            elif self._drop_indicator_position == self.BelowItem:
                # Drop Below item
                row = index.row() + 1
                col = index.column()
                index = index.parent()
        else:
            # Drop On item
            self._drop_indicator_position = self.OnViewport

        # Update given referenced list
        event_list[0], event_list[1], event_list[2], event_list[3] = event, row, col, index

        return True

    def _is_item_dropped(self, event, strict=False):
        """
        Returns whether or not an item has been dropped in given event
        :param event: QDropEvent
        :param strict: bool, True to handle ordered alphabetically list; False otherwise.
        :return: bool
        """

        is_dropped = False

        index = self.indexAt(event.pos())

        if event.source == self and event.dropAction() == Qt.MoveAction or \
                self.dragDropMode() == QAbstractItemView.InternalMove:
            top_index = QModelIndex()
            col = -1
            row = -1
            event_list = [event, row, col, top_index]
            if self._drop_on(event_list):
                event, row, col, top_index = event_list
                if row > -1:
                    if row == index.row() - 1:
                        is_dropped = False
                elif row == -1:
                    is_dropped = True
                elif row == index.row() + 1:
                    is_dropped = False if strict else True

        return is_dropped

    def _paint_drop_indicator(self, painter):
        """
        Internal function used to paint the drop indicator manually
        :param painter: QPainter
        """

        if self.state() == QAbstractItemView.DraggingState:
            opt = QStyleOption()
            opt.initFrom(self)
            opt.rect = self._drop_indicator_rect
            rect = opt.rect

            color = Qt.black
            if dcc.is_maya():
                color = Qt.white

            brush = QBrush(QColor(color))
            pen = QPen(brush, 1, Qt.DotLine)
            painter.setPen(pen)
            if rect.height() == 0:
                painter.drawLine(rect.topLeft(), rect.topRight())
            else:
                painter.drawRect(rect)

    def _edit_start(self, item):
        """
        Internal function that is called when a user start editing a tree item text
        Closes already opened edit text editors and updates internal variables
        :param item: QTreeWidgetItem
        """

        self._old_name = str(item.text(self._title_text_index))
        self.closePersistentEditor(item, self._title_text_index)
        self.openPersistentEditor(item, self._title_text_index)
        self._edit_state = item

    def _edit_finish(self, item):
        """
        Internal function that is called when a text element of the tree is edited
        Checks that that edit mode is on text and updates the item text manually
        :param item: QTreeWidgetItem
        :return: QTreeWidgetItem, edited item
        """

        if not hasattr(self._edit_state, 'text'):
            return

        self._edit_state = None

        if type(item) == int:
            return self._current_item

        self.closePersistentEditor(item, self._title_text_index)
        state = self._item_rename_valid(self._old_name, item)

        if state:
            state = self._item_renamed(item)
            if not state:
                item.setText(self._title_text_index, self._old_name)
        else:
            item.setText(self._title_text_index, self._old_name)

        return item

    def _item_rename_valid(self, old_name, item):
        """
        Checks if the rename operation on a specific item of the tree is valid or not
        :param old_name: str, old name of the item
        :param item: QTreeWidgetItem, item that is being edit
        :return: bool
        """

        new_name = item.text(self._title_text_index)
        if not new_name:
            return False

        # We do not allow duplicated names on the tree
        if self._already_exists(item):
            return False

        if old_name == new_name:
            return False

        return True

    def _already_exists(self, item):
        """
        Checks if a given QTreeWidgetItem already exists on the tree.
        :param item: QTreeWidgetItem
        :return: bool
        """

        name = item.text(0)
        parent = item.parent()

        if parent:
            skip_index = parent.indexOfChild(item)
            for i in range(parent.childCount()):
                if i == skip_index:
                    continue
                other_name = str(parent.child(i).text(0))
                if name == other_name:
                    return True
        else:
            skip_index = self.indexFromItem(item)
            skip_index = skip_index.row()
            for i in range(self.topLevelItemCount()):
                if skip_index == i:
                    continue
                other_name = str(self.topLevelItem(i).text(0))
                if name == other_name:
                    return True

        return False

    def _clear_selection(self):
        """
        Internal function used to clear the selection of the tree and update internal variables
        """

        self.clearSelection()
        self._current_item = None
        if self._edit_state:
            self._edit_finish(self._last_item)

    def _emit_item_click(self, item):
        """
        Internal function that is used to force the emission of itemClicked signal
        :param item: QTreeWidgetItem
        """

        self.itemClicked.emit(item, 0)

    def _get_ancestors(self, item):
        """
        Returns all ancestors items of the given QTreeWidgetItem
        :param item: QTreeWidgetItem
        :return: list<QTreeWidgetItem>
        """

        child_count = item.childCount()
        items = list()
        for i in range(child_count):
            child = item.child(i)
            children = self._get_ancestors(child)
            items.append(child)
            if children:
                items += children

        return items

    def _get_all_items(self):
        """
        Internal function that returns all items in the tree
        :return: list(QTreeWidgetItem)
        """

        item_count = self.topLevelItemCount()
        items = list()
        for i in range(item_count):
            item = self.topLevelItem(i)
            ancestors = self._get_ancestors(item)
            items.append(item)
            if ancestors:
                items += ancestors

        return items

    def _add_sub_items(self, tree_item):
        """
        Internal function that is updates the hiearchy of the given QTreeWidgetItem
        Implementation MUST be implemented in child class
        :param tree_item: QTreeWidgetItem
        """

        pass

    def _item_renamed(self, item):
        """
        Internal function that rename a specific QTreeWidgetItem contents
        Implementation MUST be implemented in child class
        :param item: QTreeWidgetItem
        """

        return False

    def _delete_children(self, tree_item):
        """
        Internal function that removes all children of the given item
        :param tree_item: QTreeWidgetItem
        """

        self.delete_tree_item_children(tree_item)

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_item_expanded(self, item):
        """
        Internal function that is called anytime the user expands an item of the tree
        Load dynamically all the items parented to the expanded item
        :param item: QTreeWidgetItem
        """

        if self._auto_add_sub_items:
            self._add_sub_items(item)

    def _on_item_collapsed(self, item):
        """
        Internal function that is called anytime the user collapses an item of the tree
        :param item: QTreeWidgetItem
        """

        pass

    def _on_item_activated(self, item):
        """
        Internal function that is called anytime a tree item is activated
        :param item: QTreeWidgetItem
        """

        if self._edit_state:
            self._edit_finish(self._edit_state)
        else:
            if self._text_edit:
                self._edit_start(item)

    def _on_item_changed(self, current_item, previous_item):
        """
        Internal function that is called when a tree item changes its content
        :param current_item: QTreeWidetItem
        :param previous_item: QTreeWidgetItem
        """

        if self._edit_state:
            self._edit_finish(previous_item)

    def _on_item_selection_changed(self):
        """
        Internal function that is called anytime the user changes the selection of the tree
        """

        item_sel = self.selectedItems()
        current_item = None
        if item_sel:
            current_item = item_sel[0]

        if current_item:
            self._current_name = current_item.text(self._title_text_index)
        if self._edit_state:
            self._edit_finish(self._edit_state)

        if not current_item:
            self._emit_item_click(current_item)

    def _on_item_clicked(self, item, column):
        """
        Internal function that is called anytime the user clicks on an item of the tree
        Updates internal variables and clear the selection if necessary
        :param item: QTreeWidgetItem
        :param column: int
        """

        self._last_item = self._current_item
        self._current_item = self.currentItem()
        if not item or column != self._title_text_index:
            if self._last_item:
                self._clear_selection()


class ManageTreeWidget(base.BaseWidget, object):
    def __init__(self, parent=None):
        self.tree_widget = None
        super(ManageTreeWidget, self).__init__(parent)

    def set_tree_widget(self, tree_widget):
        """
        Set the tree widget managed by this widget
        :param tree_widget: QTreeWidget
        """

        self.tree_widget = tree_widget


class FilterTreeWidget(base.DirectoryWidget, object):

    subPathChanged = Signal(str)

    def __init__(self, parent=None):
        self._tree_widget = None
        self._emit_changes = True
        self._update_tree = True
        super(FilterTreeWidget, self).__init__(parent=parent)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def get_main_layout(self):
        main_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        return main_layout

    def ui(self):
        super(FilterTreeWidget, self).ui()

        texts_layout = layouts.VerticalLayout(spacing=2, margins=(0, 0, 0, 0))
        self._filter_names = search.SearchFindWidget()
        self._filter_names.set_placeholder_text('Filter Names')
        self._sub_path_filter = lineedit.BaseLineEdit()
        self._sub_path_filter.setPlaceholderText('Set Sub Path')
        self._sub_path_filter.setVisible(False)
        texts_layout.addWidget(self._filter_names)
        texts_layout.addWidget(self._sub_path_filter)
        self.main_layout.addLayout(texts_layout)

    def setup_signals(self):
        self._filter_names.textChanged.connect(self._on_filter_names)
        self._sub_path_filter.textChanged.connect(self._on_sub_path_filter_changed)
        self._sub_path_filter.textEdited.connect(self._on_sub_path_filter_edited)

    def dropEvent(self, event):
        item = self.item_at(event.pos())
        if item:
            item_path = self.get_tree_item_path_string(item)
            item_full_path = path.join_path(self._directory, item_path)
            if not item_full_path or not path.is_dir(item_full_path):
                event.ignore()
            else:
                event.accept()
        else:
            event.ignore()

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def get_name_filter(self):
        """
        Returns the name filter current text
        :return: str
        """

        return str(self._filter_names.text())

    def get_sub_path_filter(self):
        """
        Returns the name of the sub path filters text
        :return: str
        """

        return str(self._sub_path_filter.text())

    def set_emit_changes(self, flag):
        """
        Sets whether signals should be emitted or not
        :param flag: bool
        """

        self._emit_changes = flag

    def set_name_filter(self, text):
        """
        Sets the name filter text
        :param text: str
        """

        self._filter_names.setText(text)

    def set_sub_path_filter(self, text):
        """
        Sets the name sub path filter text
        :param text: str
        """

        self._sub_path_filter.setText(text)

    def clear_name_filter(self):
        """
        Clears current name filter text
        """

        self._filter_names.setText('')

    def clear_sub_path_filter(self):
        """
        Clears current sub path filter text
        """

        self._sub_path_filter.setText('')

    def set_tree_widget(self, tree_widget):
        """
        Sets the tree widget used by this widget
        :param tree_widget: QTreeWidget
        """

        self._tree_widget = tree_widget

    def set_sub_path_warning(self, flag):
        """
        Sets whether or not sub path filter text should indicate a warning
        :param flag: bool
        """

        if flag:
            if dcc.is_maya():
                self._sub_path_filter.setStyleSheet('background-color: rgb(255, 100, 100);')
            else:
                self._sub_path_filter.setStyleSheet('background-color: rgb(255, 150, 150);')
        else:
            self._sub_path_filter.setStyleSheet('')

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_filter_names(self, text):
        """
        Internal callback function that is used to call the filter function of the TreeWidget
        :param text: str, filter text that should be used
        """

        if self._update_tree:
            self._tree_widget.filter_names(text)

    def _on_sub_path_filter_changed(self):
        """
        Internal callback function that is called when sub path filter text changes
        """

        current_text = str(self._sub_path_filter.text()).strip()
        self.subPathChanged.emit(current_text)

    def _on_sub_path_filter_edited(self):
        """
        Internal callback function that is called when sub path filter text is edited
        """

        current_text = str(self._sub_path_filter.text()).strip()
        if not current_text:
            self.set_directory(self._directory)
            if self._update_tree:
                self._tree_widget.set_directory(self._directory)
            text = self._filter_names.text
            self._on_filter_names(text)
            return

        sub_dir = path.join_path(self._directory, current_text)
        if not sub_dir:
            return

        if path.is_dir(sub_dir):
            if self._update_tree:
                self._tree_widget.set_directory(self._directory)
            text = self._filter_names.text
            self._on_filter_names(text)

        if self._emit_changes:
            self.subPathChanged.emit(current_text)


class FileTreeWidget(TreeWidget, object):

    refreshed = Signal()

    HEADER_LABELS = ['Name', 'Size MB', 'Time']
    NEW_ITEM_NAME = 'new_file'
    ITEM_WIDGET = QTreeWidgetItem
    EXCLUDE_EXTENSIONS = list()

    def __init__(self, parent=None):
        self._directory = None
        super(FileTreeWidget, self).__init__(parent)

        self.setHeaderLabels(self.HEADER_LABELS)

    # ============================================================================================================
    # PROPERTIES
    # ============================================================================================================

    @property
    def directory(self):
        return self._directory

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def dropEvent(self, event):
        item = self.item_at(event.pos())
        if item:
            item_path = self.get_tree_item_path_string(item)
            item_full_path = path.join_path(self._directory, item_path)
            if item_full_path and path.is_dir(item_full_path):
                super(FileTreeWidget, self).dropEvent(event)

    def _add_item(self, file_name, parent=None):
        """
        Function that adds given file into the tree
        :param file_name: str, name of the file new item will store
        :param parent: QTreeWidgetItem, parent item to append new item into
        :return: QTreeWidet, new item added
        """

        try:
            self.blockSignals(True)
            self.clearSelection()
        finally:
            self.blockSignals(False)

        path_name = file_name
        found = False

        # Check if item exists
        if parent:
            parent_path = self.get_tree_item_path_string(parent)
            path_name = '{}/{}'.format(parent_path, file_name)
            for i in range(parent.childCount()):
                item = parent.child(i)
                if item.text(0) == file_name:
                    found = item
        else:
            for i in range(self.topLevelItemCount()):
                item = self.topLevelItem(i)
                if item.text(0) == file_name:
                    found = item

        # Check if the item should be excluded or not from the tree
        exclude = self.EXCLUDE_EXTENSIONS
        if exclude:
            split_name = file_name.split('.')
            extension = split_name[-1]
            if extension in exclude:
                return

        if found:
            item = found
        else:
            item = self.create_item_widget(file_name)

        # Constrain item size if necessary
        size = self.ITEM_WIDGET_SIZE
        if size:
            size = QSize(*size)
            item.setSizeHint(self._title_text_index, size)

        # Set item text
        item_path = path.join_path(self._directory, path_name)
        sub_files = folder.get_files_and_folders(item_path)
        item.setText(self._title_text_index, file_name)

        # Retrieve file properties
        if self.header().count() > 1:
            if path.is_file(item_path):
                size = fileio.get_file_size(item_path)
                date = fileio.get_last_modified_date(item_path)
                item.setText(self._title_text_index + 1, str(size))
                item.setText(self._title_text_index + 2, str(date))

        # Update valid sub files
        # NOTE: Sub files are added dynamically when the user expands an item
        if sub_files:
            self._delete_children(item)
            exclude_extensions = self.EXCLUDE_EXTENSIONS
            exclude_count = 0
            if exclude_extensions:
                for f in sub_files:
                    for exclude in exclude_extensions:
                        if f.endswith(exclude):
                            exclude_count += 1
                            break

            if exclude_count != len(sub_files):
                QTreeWidgetItem(item)

        # Add item to tree hierarchy
        if parent:
            parent.addChild(item)
            try:
                self.blockSignals(True)
                self.setCurrentItem(item)
            finally:
                self.blockSignals(False)
        else:
            self.addTopLevelItem(item)

        return item

    def _add_items(self, files, parent=None):
        """
        Function that adds given files into the tree
        :param files: list<str>, list of files to add
        :param parent: QTreeWidgetItem, parent item to append new item
        :return:
        """

        if not files:
            return

        for filename in files:
            if parent:
                self._add_item(filename, parent)
            else:
                self._add_item(filename)

    def _add_sub_items(self, tree_item):
        """
        Implements _add_sub_items() functionality
        :param tree_item: QTreeWidgetItem
        """

        # Clean item hierarchy first
        self.delete_empty_children(tree_item)
        self._delete_children(tree_item)

        path_str = self.get_tree_item_path_string(tree_item)
        full_path_str = path.join_path(self._directory, path_str)
        files = self._get_files(full_path_str)

        self._add_items(files, tree_item)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def set_directory(self, directory, refresh=True, name_filter=None):
        """
        Sets the directory used by this QTreeWidget
        :param directory: str, directory
        :param refresh: bool, Whether to refresh QTreeWidget items after setting working directory
        """

        self._directory = directory
        self._name_filter = name_filter
        if refresh:
            self.refresh()

    def get_item_directory(self, tree_item):
        """
        Returns the full path of the given tree item
        :param tree_item: QTreeWidgetItem
        :return: str
        """

        path_str = self.get_tree_item_path_string(tree_item)
        return path.join_path(self._directory, path_str)

    def create_item_widget(self, file_name):
        """
        Creates a new item widget
        :return: variant
        """

        return self.ITEM_WIDGET()

    def create_item(self, name=None):
        """
        Creates a new item (folder) inside the selected item
        :param name: str, name of the new item
        """

        current_item = self._current_item
        if current_item:
            item_path = self.get_tree_item_path_string(self._current_item)
            item_path = path.join_path(self._directory, item_path)
            if path.is_file(item_path):
                item_path = path.get_dirname(item_path)
                current_item = self._current_item.parent()

        if not current_item:
            item_path = self._directory

        if not name:
            name = self.NEW_ITEM_NAME

        folder.create_folder(name=name, directory=item_path)

        if current_item:
            self._add_sub_items(current_item)
            self.setItemExpanded(current_item, True)
        else:
            self.refresh()

    def delete_item(self):
        """
        Deletes the selected item
        """

        item = self._current_item
        item_path = self.get_item_directory(item)
        name = path.get_basename(item_path)
        item_directory = path.get_dirname(item_path)

        if path.is_dir(item_path):
            folder.delete_folder(name, item_directory)
        elif path.is_file(item_path):
            fileio.delete_file(name, item_directory)
            if item_path.endswith('.py'):
                fileio.delete_file(name + '.c', item_directory)

        parent = item.parent()
        if parent:
            parent.removeChild(item)
        else:
            index = self.indexOfTopLevelItem(item)
            self.takeTopLevelItem(index)

    def refresh(self):
        """
        Refreshes all QTreeWidget items
        """

        if not self._directory:
            self.clear()
            return

        files = self._get_files()
        if not files:
            self.clear()
            return

        self._load_files(files)
        self.refreshed.emit()

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _get_files(self, directory=None):
        """
        Internal function taht returns  all files located in the given directory. If not directory is given, stored
         variable directory will be used
        :return: list<str>
        """

        if not directory:
            directory = self._directory

        return folder.get_files_and_folders(directory)

    def _load_files(self, files):
        """
        Internal function that adds given files into the tree (clearing the tree first)
        :param files: list<str>
        """

        self.clear()
        self._add_items(files)


class EditFileTreeWidget(base.DirectoryWidget, object):

    itemClicked = Signal(object, object)
    description = 'EditTree'

    TREE_WIDGET = FileTreeWidget
    MANAGER_WIDGET = ManageTreeWidget
    FILTER_WIDGET = FilterTreeWidget

    def __init__(self, parent=None):
        super(EditFileTreeWidget, self).__init__(parent)

        self._on_edit(False)

    # ============================================================================================================
    # PROPERTIES
    # ============================================================================================================

    @property
    def tree_widget(self):
        return self._tree_widget

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def ui(self):
        super(EditFileTreeWidget, self).ui()

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self._tree_widget = self.TREE_WIDGET(parent=self)

        self._manager_widget = self.MANAGER_WIDGET(parent=self)
        self._manager_widget.set_tree_widget(self._tree_widget)

        self._filter_widget = self.FILTER_WIDGET(parent=self)
        self._filter_widget.set_tree_widget(self._tree_widget)
        self._filter_widget.set_directory(self._directory)
        drag_reorder_icon = resources.icon('drag_reorder')
        edit_mode_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        edit_mode_layout.setAlignment(Qt.AlignBottom)
        self._edit_mode_btn = buttons.BaseButton(parent=self)
        self._edit_mode_btn.setIcon(drag_reorder_icon)
        self._edit_mode_btn.setCheckable(True)
        edit_mode_layout.addWidget(self._edit_mode_btn)
        self._filter_widget.main_layout.addLayout(edit_mode_layout)

        self.main_layout.addWidget(self._filter_widget)
        self.main_layout.addWidget(self._tree_widget)
        self.main_layout.addWidget(self._manager_widget)

    def setup_signals(self):
        self._edit_mode_btn.toggled.connect(self._on_edit)
        self._tree_widget.itemClicked.connect(self._on_item_selection_changed)

    def set_directory(self, directory, refresh=True):
        """
        Overrides set_directory function to take in account also the sub path
        and update all directories of the different tree widgets
        :param directory: str
        """

        super(EditFileTreeWidget, self).set_directory(directory)
        self._filter_widget.set_directory(directory)
        self._tree_widget.set_directory(directory, refresh=refresh)

        if hasattr(self._manager_widget, 'set_directory'):
            self._manager_widget.set_directory(directory)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def get_current_item(self):
        """
        Returns the current selected item on the tree
        :return: TreeItem
        """

        items = self._tree_widget.selectedItems()
        item = None
        if items:
            item = items[0]

        return item

    def get_current_item_name(self):
        """
        Returns the current name of the selected item on the tree
        :return: str
        """

        current_item = self.get_current_item()
        return current_item.text(0) if current_item else ''

    def get_current_item_directory(self):
        """
        Returns the directory the current selected item points to
        :return: str
        """

        item = self.get_current_item()
        return self._tree_widget.get_item_directory(item)

    def enable_edit_mode(self):
        """
        Enables edit mode
        """

        if not self._edit_mode_btn.isChecked():
            self._edit_mode_btn.blockSignals(True)
            try:
                self._edit_mode_btn.setChecked(True)
            finally:
                self._edit_mode_btn.blockSignals(False)

        self._tree_widget.setDragEnabled(True)
        self._tree_widget.setAcceptDrops(True)
        self._tree_widget.setDropIndicatorShown(True)

    def disable_edit_mode(self):
        """
        Enables edit mode
        """

        if self._edit_mode_btn.isChecked():
            self._edit_mode_btn.blockSignals(True)
            try:
                self._edit_mode_btn.setChecked(False)
            finally:
                self._edit_mode_btn.blockSignals(False)

        self._tree_widget.setDragEnabled(False)
        self._tree_widget.setAcceptDrops(False)
        self._tree_widget.setDropIndicatorShown(False)

    def get_checked_children(self, tree_item):
        """
        Function that returns checked item children of the given tree item
        :param tree_item:
        :return:
        """

        if not tree_item:
            return

        expand_state = tree_item.isExpanded()
        tree_item.setExpanded(True)
        children = self._tree_widget.get_tree_item_children(tree_item)

        checked_children = list()
        for child in children:
            check_state = child.checkState(0)
            if check_state == Qt.Checked:
                checked_children.append(child)
        levels = list()
        if checked_children:
            levels.append(checked_children)

        while children:
            new_children = list()
            checked_children = list()
            for child in children:
                current_check_state = child.checkState(0)
                if current_check_state != Qt.Checked:
                    continue
                child.setExpanded(True)
                sub_children = self._tree_widget.get_tree_item_children(child)
                checked = list()
                for sub_child in sub_children:
                    check_state = sub_child.checkState(0)
                    if check_state == Qt.Checked:
                        checked.append(sub_child)
                if sub_children:
                    new_children += sub_children
                if checked:
                    checked_children += checked
            if not checked_children:
                children = list()
                continue
            children = new_children
            if checked_children:
                levels.append(checked_children)

        tree_item.setExpanded(expand_state)
        levels.reverse()

        return levels

    def refresh(self):
        """
        Refresh TreeWidget items
        """

        self._tree_widget.refresh()

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_item_selection_changed(self):
        """
        Internal function that is called anytime the user selects an item on the TreeWidget
        Emits itemClicked signal with the name of the selected item and the item itself
        """

        items = self._tree_widget.selectedItems()
        name = None
        item = None
        if items:
            item = items[0]
            name = item.text(0)
            self.itemClicked.emit(name, item)

        return name, item

    def _on_edit(self, flag):
        """
        Internal function that is called anytime the user presses the Edit button on the filter widget
        If edit is ON, drag/drop operations in tree widget are disabled
        :param flag: bool
        """

        if flag:
            self.enable_edit_mode()
        else:
            self.disable_edit_mode()


class TreeWidgetItem(QTreeWidgetItem, object):
    def __init__(self, parent=None):
        self._widget = self._get_widget()
        if self._widget:
            self._widget.item_view = self
        self.column = self._get_column()
        super(TreeWidgetItem, self).__init__(parent)

    def _get_widget(self):
        return None

    def _get_column(self):
        return 0



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains examples of toolbar classes
"""

from __future__ import print_function, division, absolute_import

from Qt.QtWidgets import QWidget, QAction
from Qt.QtGui import QColor

from tpDcc.managers import resources
from tpDcc.libs.qt.core import base, contexts
from tpDcc.libs.qt.widgets import layouts, stack, lineedit, toolbar, buttons


class ToolbarExampleWidget(base.BaseWidget):

    def ui(self):
        super(ToolbarExampleWidget, self).ui()

        self._stack = stack.SlidingOpacityStackedWidget(parent=self)
        self.main_layout.addWidget(self._stack)

        for widget in [self._toolbar_example()]:
            self._stack.addWidget(widget)

    def _toolbar_example(self):

        def _collapse():
            tool_bar.collapse()

        def _set_icon_color():
            tool_bar.set_icon_color(QColor(255, 255, 0))

        toolbar_widget = QWidget(parent=self)
        toolbar_layout = layouts.VerticalLayout()
        toolbar_widget.setLayout(toolbar_layout)

        tool_bar = toolbar.ToolBar(parent=self)
        tool_bar.set_children_height(50)
        toolbar_layout.addWidget(tool_bar)
        line_edit = lineedit.BaseLineEdit(parent=self)
        line_edit.setObjectName('hello')

        collapse_action = tool_bar.addAction('Collapse')
        tool_bar.addWidget(line_edit)
        tool_bar.addAction(resources.icon('add'), 'Plus')
        color_action = QAction('Yellow', None)
        tool_bar.insertAction('Plus', color_action)

        collapse_action.triggered.connect(_collapse)
        color_action.triggered.connect(_set_icon_color)

        self.main_layout.addWidget(buttons.BaseButton('Hello'))
        self.main_layout.addStretch()

        return toolbar_widget


if __name__ == '__main__':
    import tpDcc.loader
    tpDcc.loader.init()

    with contexts.application():
        widget = ToolbarExampleWidget()
        with contexts.show_window(widget):
            pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains widgets related with search functionality
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Signal, QSize, QEvent, QPoint
from Qt.QtWidgets import QApplication, QWidget, QLineEdit, QStyle, QMenu
from Qt.QtGui import QIcon

from tpDcc.managers import resources
from tpDcc.libs.qt.widgets import layouts, buttons


class SearchFindWidget(QWidget, object):

    textChanged = Signal(str)
    editingFinished = Signal(str)
    returnPressed = Signal()

    def __init__(self, search_line=None, parent=None):
        super(SearchFindWidget, self).__init__(parent=parent)

        self.setObjectName('SearchFindWidget')

        self.text = ''
        self._placeholder_text = ''

        main_layout = layouts.HorizontalLayout(spacing=2, margins=(2, 2, 2, 2))
        self.setLayout(main_layout)

        self._search_line = search_line or QLineEdit(self)
        self._search_menu = QMenu()
        self._search_menu.addAction('Test')

        icon_size = self.style().pixelMetric(QStyle.PM_SmallIconSize)

        delete_icon = resources.icon('delete')
        search_icon = QIcon(resources.icon('search'))

        self._clear_btn = buttons.IconButton(delete_icon, icon_padding=2, parent=self)
        self._clear_btn.setIconSize(QSize(icon_size, icon_size))
        self._clear_btn.setFixedSize(QSize(icon_size, icon_size))
        self._clear_btn.hide()

        self._search_btn = buttons.IconButton(search_icon, icon_padding=2, parent=self)
        self._search_btn.setIconSize(QSize(icon_size, icon_size))
        self._search_btn.setFixedSize(QSize(icon_size, icon_size))
        # self._search_btn.setStyleSheet('border: none;')
        # self._search_btn.setPopupMode(QToolButton.InstantPopup)
        self._search_btn.setEnabled(True)

        self._search_line.setStyleSheet(
            """
            QLineEdit { padding-left: %spx; padding-right: %spx; border-radius:10px; border:2px; border-color:red; }
            """ % (self._search_button_padded_width(), self._clear_button_padded_width())
        )
        self._search_line.setMinimumSize(
            max(
                self._search_line.minimumSizeHint().width(),
                self._clear_button_padded_width() + self._search_button_padded_width()),
            max(
                self._search_line.minimumSizeHint().height(),
                max(self._clear_button_padded_width(), self._search_button_padded_width()))
        )

        main_layout.addWidget(self._search_line)

        self._search_line.setFocus()

        self._search_line.textChanged.connect(self.textChanged)
        self._search_line.textChanged.connect(self.set_text)
        # self._search_line.editingFinished.connect(self.editingFinished)
        # self._search_line.returnPressed.connect(self.returnPressed)
        self._clear_btn.clicked.connect(self.clear)
        self._search_btn.clicked.connect(self._popup_menu)

    @property
    def search_line(self):
        return self._search_line

    def changeEvent(self, event):
        if event.type() == QEvent.EnabledChange:
            enabled = self.isEnabled()
            self._search_btn.setEnabled(enabled and self._search_menu)
            self._search_line.setEnabled(enabled)
            self._clear_btn.setEnabled(enabled)
        super(SearchFindWidget, self).changeEvent(event)

    def resizeEvent(self, event):
        if not (self._clear_btn and self._search_line):
            return
        super(SearchFindWidget, self).resizeEvent(event)
        x = self.width() - self._clear_button_padded_width() * 0.85
        y = (self.height() - self._clear_btn.height()) * 0.5
        self._clear_btn.move(x - 3, y)
        self._search_btn.move(self._search_line_frame_width() * 2, (self.height() - self._search_btn.height()) * 0.5)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.clear()
        super(SearchFindWidget, self).keyPressEvent(event)

    def get_text(self):
        if not self._search_line:
            return ''
        return self._search_line.text()

    def set_text(self, text):
        if not (self._clear_btn and self._search_line):
            return

        self._clear_btn.setVisible(not (len(text) == 0))
        if text != self.get_text():
            self._search_line.setText(text)

    def get_placeholder_text(self):
        if not self._search_line:
            return ''

        return self._search_line.text()

    def set_placeholder_text(self, text):
        if not self._search_line:
            return
        self._search_line.setPlaceholderText(text)

    def get_menu(self):
        search_icon = resources.icon('search')
        self._search_btn.setIcon(search_icon)
        self._search_btn.setEnabled(self.isEnabled() and self._menu)

    def set_focus(self, reason=Qt.OtherFocusReason):
        if self._search_line:
            self._search_line.setFocus(reason)
        else:
            self.setFocus(Qt.OtherFocusReason)

    def clear(self):
        if not self._search_line:
            return
        self._search_line.clear()
        self.set_focus()

    def select_all(self):
        if not self._search_line:
            return
        self._search_line.selectAll()

    def _search_line_frame_width(self):
        return self._search_line.style().pixelMetric(QStyle.PM_DefaultFrameWidth)

    def _clear_button_padded_width(self):
        return self._clear_btn.width() + self._search_line_frame_width() * 2

    def _clear_button_padded_height(self):
        return self._clear_btn.height() + self._search_line_frame_width() * 2

    def _search_button_padded_width(self):
        return self._search_btn.width() + self._search_line_frame_width() * 2

    def _search_button_padded_height(self):
        return self._search_btn.height() + self._search_line_frame_width() * 2

    def _popup_menu(self):
        if self._search_menu:
            screen_rect = QApplication.desktop().availableGeometry(self._search_btn)
            size_hint = self._search_menu.sizeHint()
            rect = self._search_btn.rect()
            top_diff = rect.top() - size_hint.height()
            x = rect.right() - size_hint.width() if self._search_btn.isRightToLeft() else rect.left()
            y = rect.bottom() if self._search_btn.mapToGlobal(
                QPoint(0, rect.bottom())).y() + size_hint.height() <= screen_rect.height() else top_diff
            point = self._search_btn.mapToGlobal(QPoint(x, y))
            point.setX(max(screen_rect.left(), min(point.x(), screen_rect.right() - size_hint.width())))
            point.setY(point.y() + 1)
            print('pop up on {}'.format(point))
            self._search_menu.popup(point)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains widgets related with directories and files
"""

from __future__ import print_function, division, absolute_import

import os
import logging

from tpDcc import dcc
from Qt.QtCore import Qt, Signal
from Qt.QtWidgets import QSizePolicy, QWidget, QListWidget, QAbstractItemView
from Qt.QtGui import QColor, QPalette

from tpDcc.managers import resources
from tpDcc.libs.python import path
from tpDcc.libs.qt.core import base, qtutils
from tpDcc.libs.qt.widgets import layouts, buttons, lineedit, label

LOGGER = logging.getLogger('tpDcc-libs-qt')


class FileListWidget(QListWidget, object):
    """
    Widgets that shows files and directories such Windows Explorer
    """

    directory_activated = Signal(str)
    file_activated = Signal(str)
    file_selected = Signal(str)
    folder_selected = Signal(str)
    directory_selected = Signal(str)
    files_selected = Signal(list)
    up_requested = Signal()
    update_requested = Signal()

    def __init__(self, parent):
        self.parent = parent
        super(FileListWidget, self).__init__(parent)

        self.itemSelectionChanged.connect(self.selectItem)
        self.itemDoubleClicked.connect(self.activateItem)

    def resizeEvent(self, event):
        """
        Overrides QWidget resizeEvent so when the widget is resize a update request signal is emitted
        :param event: QResizeEvent
        """

        self.update_requested.emit()
        super(FileListWidget, self).resizeEvent(event)

    def wheelEvent(self, event):
        """
        Overrides QWidget wheelEvent to smooth scroll bar movement
        :param event: QWheelEvent
        """

        sb = self.horizontalScrollBar()
        minValue = sb.minimum()
        maxValue = sb.maximum()
        if sb.isVisible() and maxValue > minValue:
            sb.setValue(sb.value() + (-1 if event.delta() > 0 else 1))
        super(FileListWidget, self).wheelEvent(event)

    def keyPressEvent(self, event):
        """
        Overrides QWidget keyPressEvent with some shortcuts when using the widget
        :param event:
        :return:
        """
        modifiers = event.modifiers()
        if event.key() == int(Qt.Key_Return) and modifiers == Qt.NoModifier:
            if len(self.selectedItems()) > 0:
                item = self.selectedItems()[0]
                if item.type() == 0:  # directory
                    self.directory_activated.emit(item.text())
                else:  # file
                    self.file_activated.emit(item.text())
        elif event.key() == int(Qt.Key_Backspace) and modifiers == Qt.NoModifier:
            self.up_requested.emit()
        elif event.key() == int(Qt.Key_F5) and modifiers == Qt.NoModifier:
            self.update_requested.emit()
        else:
            super(FileListWidget, self).keyPressEvent(event)

    def selectItem(self):
        if len(self.selectedItems()) > 0:
            item = self.selectedItems()[0]
            if item.type() == 0:    # directory
                self.folder_selected.emit(item.text())
            if item.type() == 1:  # file
                self.file_selected.emit(item.text())

    def activateItem(self):
        if len(self.selectedItems()) > 0:
            item = self.selectedItems()[0]
            if item.type() == 0:  # directory
                self.directory_activated.emit(item.text())
            else:  # file
                self.file_activated.emit(item.text())

    def setExtendedSelection(self):
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.itemSelectionChanged.disconnect(self.selectItem)
        self.itemSelectionChanged.connect(self.processSelectionChanged)

    def processSelectionChanged(self):
        """
        Gets all selected items and emits a proper signal with the proper selected item names
        """

        items = filter(lambda x: x.type() != 0, self.selectedItems())
        names = map(lambda x: x.text(), items)
        self.files_selected.emit(names)


class FolderEditLine(lineedit.BaseLineEdit, object):
    """
    Custom QLineEdit with drag and drop behaviour for files and folders
    """

    def __init__(self, parent=None):
        super(FolderEditLine, self).__init__(parent)

        self.setDragEnabled(True)
        self.setReadOnly(True)

    def dragEnterEvent(self, event):
        """
        Overrides QWidget dragEnterEvent to enable drop behaviour with file
        :param event: QDragEnterEvent
        :return:
        """
        data = event.mimeData()
        urls = data.urls()
        if (urls and urls[0].scheme() == 'file'):
            event.acceptProposedAction()

    def dragMoveEvent(self, event):
        data = event.mimeData()
        urls = data.urls()
        if (urls and urls[0].scheme() == 'file'):
            event.acceptProposedAction()

    def dropEvent(self, event):
        data = event.mimeData()
        urls = data.urls()
        if (urls and urls[0].scheme() == 'file'):
            self.setText(urls[0].toLocalFile())


class SelectFolderButton(QWidget, object):
    """
    Button widget that allows to select folder paths
    """

    beforeNewDirectory = Signal()
    directoryChanged = Signal(object)   # Signal that is called when a new folder is selected

    def __init__(self, text='Browse', directory='', use_app_browser=False, parent=None):
        super(SelectFolderButton, self).__init__(parent)

        self._use_app_browser = use_app_browser
        self._directory = directory
        self.settings = None

        main_layout = layouts.HorizontalLayout(spacing=2, margins=(2, 2, 2, 2))
        self.setLayout(main_layout)

        folder_icon = resources.icon('folder')
        self._folder_btn = buttons.BaseButton(text, parent=self)
        self._folder_btn.setIcon(folder_icon)
        main_layout.addWidget(self._folder_btn)

        self._folder_btn.clicked.connect(self._open_folder_browser_dialog)

    @property
    def folder_btn(self):
        return self._folder_btn

    def get_init_directory(self):
        return self._directory

    def set_init_directory(self, directory):
        self._directory = directory

    init_directory = property(get_init_directory, set_init_directory)

    def set_settings(self, settings):
        self.settings = settings

    def _open_folder_browser_dialog(self):
        """
        Opens a set folder browser and returns the selected path
        :return: str, Path of the selected folder
        """

        self.beforeNewDirectory.emit()

        result = dcc.select_folder_dialog('Select Folder', start_directory=self.init_directory) or ''

        self.directoryChanged.emit(result)
        # if not result or not os.path.isdir(result[0]):
        if not result or not os.path.isdir(result):
            return
        return path.clean_path(result[0])


class SelectFolder(QWidget, object):
    """
    Widget with button and line edit that opens a folder dialog to select folder paths
    """

    directoryChanged = Signal(object)  # Signal that is called when a new folder is selected

    def __init__(self, label_text='Select Folder', directory='', use_app_browser=False, use_icon=True, parent=None):
        super(SelectFolder, self).__init__(parent)

        self._use_app_browser = use_app_browser
        self._use_icon = use_icon
        self.settings = None
        self.directory = None
        self._label_text = label_text
        self._directory = directory

        main_layout = layouts.HorizontalLayout(spacing=2, margins=(2, 2, 2, 2))
        self.setLayout(main_layout)

        self._folder_label = label.BaseLabel(
            '{0}'.format(self._label_text)) if self._label_text == '' else label.BaseLabel(
            '{0}:'.format(self._label_text))
        if not self._label_text:
            self._folder_label.setVisible(False)
        self._folder_line = FolderEditLine()
        self._folder_line.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        if os.path.exists(self._directory):
            self._folder_line.setText(self._directory)

        self._folder_btn = buttons.BaseButton(parent=self)
        if self._use_icon:
            folder_icon = resources.icon('folder')
            self._folder_btn.setIcon(folder_icon)
        else:
            self._folder_btn.setText('Browse ...')

        for widget in [self._folder_label, self._folder_line, self._folder_btn]:
            main_layout.addWidget(widget)

        self._folder_btn.clicked.connect(self._open_folder_browser_dialog)

    @property
    def folder_label(self):
        return self._folder_label

    @property
    def folder_line(self):
        return self._folder_line

    @property
    def folder_btn(self):
        return self._folder_btn

    def set_directory_text(self, new_text):
        """
        Sets the text of the directory line
        :param new_text: str
        """

        self._folder_line.setText(new_text)

    def get_directory(self):
        """
        Returns directory set on the directory line
        :return: str
        """

        return str(self._folder_line.text())

    def set_directory(self, directory):
        """
        Sets the directory of the directory line
        """

        if not directory:
            return

        self.directory = directory

        self.set_directory_text(directory)

    def _open_folder_browser_dialog(self):
        """
        Opens a set folder browser and returns the selected path
        :return: str, Path of the selected folder
        """

        result = dcc.select_folder_dialog('Select Folder', start_directory=self.folder_line.text()) or ''
        if not result or not os.path.isdir(result):
            return
        else:
            filename = path.clean_path(result)
            self.set_directory(filename)
            self._text_changed()

        return filename

    def _text_changed(self):
        """
        This function is called each time the user manually changes the line text
        Emits the signal to notify that the directory has changed
        :param directory: str, new edit line text after user edit
        """

        directory = self.get_directory()
        if path.is_dir(directory):
            self.directoryChanged.emit(directory)

    def _send_directories(self, directory):
        """
        Emit the directory changed signal with the given directory
        :param directory: str
        :return: str
        """

        self.directoryChanged.emit(directory)


class SelectFile(base.DirectoryWidget, object):
    """
    Widget with button and line edit that opens a file dialog to select file paths
    """

    directoryChanged = Signal(object)  # Signal that is called when a new folder is selected

    def __init__(self, label_text='Select File', directory='', use_app_browser=False,
                 filters=None, use_icon=True, parent=None):

        self._use_app_browser = use_app_browser
        self.settings = None
        self._use_icon = use_icon
        self._directory = directory
        self._label_text = label_text
        self._filters = filters

        super(SelectFile, self).__init__(parent)

    @property
    def file_label(self):
        return self._file_label

    @property
    def file_line(self):
        return self._file_line

    @property
    def file_btn(self):
        return self._folder_btn

    def get_main_layout(self):
        main_layout = layouts.HorizontalLayout(spacing=2, margins=(2, 2, 2, 2))
        return main_layout

    def ui(self):
        super(SelectFile, self).ui()

        self._file_label = label.BaseLabel(
            '{0}'.format(self._label_text)) if self._label_text == '' else label.BaseLabel(
            '{0}:'.format(self._label_text), parent=self)
        if not self._label_text:
            self._file_label.setVisible(False)
        self._file_line = FolderEditLine()
        self._file_line.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        if self._directory and os.path.exists(self._directory):
            self._file_line.setText(self._directory)

        self._file_btn = buttons.BaseButton(parent=self)
        self._clear_btn = buttons.BaseButton(parent=self)
        if self._use_icon:
            self._file_btn.setIcon(resources.icon('folder'))
            self._clear_btn.setIcon(resources.icon('delete'))
        else:
            self._file_btn.setText('Browse ...')
            self._clear_btn.setText('Clear')

        for widget in [self._file_label, self._file_line, self._file_btn, self._clear_btn]:
            self.main_layout.addWidget(widget)

    def setup_signals(self):
        self._file_btn.clicked.connect(self._open_file_browser_dialog)
        self._clear_btn.clicked.connect(self._on_reset_path)
        self._file_line.textChanged.connect(self._text_changed)

    def set_settings(self, settings):
        """
        Set new settings. Override in new classes to add custom behaviour
        :param settings:
        :return:
        """
        self.settings = settings

    def update_settings(self, filename):
        """
        Updates current settings. Override in new classes to add custom behaviour
        :param settings: new selected path for the user
        """

        pass

    def set_label(self, text):
        """
        Sets the directory label text
        :param text: str, new directory label text
        :return:
        """

        self._file_label.setText(text)
        self._file_label.setVisible(bool(text))

    def set_directory(self, directory):
        """
        Sets the text of the directory line
        :param directory: str
        """

        super(SelectFile, self).set_directory(directory=directory)

        self._file_line.setText(directory)

    def get_directory(self):
        """
        Returns directory setted on the directory line
        :return: str
        """

        return self._file_line.text()

    def _open_file_browser_dialog(self):
        """
        Opens a set folder browser and returns the selected path
        :return: str, Path of the selected folder
        """

        file_line = self._file_line.text()
        if os.path.isfile(file_line):
            file_line = os.path.dirname(file_line)
        result = dcc.select_file_dialog('Select File', start_directory=file_line, pattern=self._filters or '') or ''
        if not result or not os.path.isfile(result):
            LOGGER.warning('Selected file "{}" is not a valid file!'.format(result))
            return
        else:
            filename = path.clean_path(result)
            self.set_directory(filename)
            self.directoryChanged.emit(filename)
            self.update_settings(filename=filename)

        return filename

    def _on_reset_path(self):
        self.set_directory('')
        self.directoryChanged.emit('')
        self.update_settings(filename='')

    def _text_changed(self):
        """
        This function is called each time the user manually changes the line text
        :param directory: str, new edit line text after user edit
        """

        f = self.get_directory()
        if path.is_file(f):
            self.directoryChanged.emit(f)


class GetDirectoryWidget(base.DirectoryWidget, object):
    directoryChanged = Signal(object)
    textChanged = Signal(object)

    def __init__(self, parent=None):

        self._label = 'directory'
        self._show_files = False

        super(GetDirectoryWidget, self).__init__(parent=parent)

    @property
    def show_files(self):
        return self._show_files

    def ui(self):
        super(GetDirectoryWidget, self).ui()

        self._directory_widget = QWidget()
        directory_layout = layouts.HorizontalLayout()
        self._directory_widget.setLayout(directory_layout)
        self.main_layout.addWidget(self._directory_widget)

        self._directory_lbl = label.BaseLabel('directory', parent=self)
        self._directory_lbl.setMinimumWidth(60)
        self._directory_lbl.setMaximumWidth(100)
        self._directory_edit = lineedit.BaseLineEdit(parent=self)
        self._directory_edit.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self._directory_browse_btn = buttons.BaseButton('browse', parent=self)

        directory_layout.addWidget(self._directory_lbl)
        directory_layout.addWidget(self._directory_edit)
        directory_layout.addWidget(self._directory_browse_btn)

    def setup_signals(self):
        self._directory_edit.textChanged.connect(self._on_text_changed)
        self._directory_browse_btn.clicked.connect(self._on_browse)

    def set_directory(self, directory):
        super(GetDirectoryWidget, self).set_directory(directory)
        self.set_directory_text(directory)

    def get_directory(self):
        return self._directory_edit.text()

    def set_label(self, label):
        length = len(label) * 8
        self._directory_lbl.setMinimumWidth(length)
        self._directory_lbl.setText(label)

    def set_directory_text(self, text):
        self._directory_edit.setText(text)

    def set_placeholder_text(self, text):
        self._directory_edit.setPlaceholderText(text)

    def set_example(self, text):
        self.set_placeholder_text('example: {}'.format(text))

    def set_error(self, flag):

        if dcc.is_maya():
            yes_color = QColor(0, 255, 0, 50)
            no_color = QColor(255, 0, 0, 50)
        else:
            yes_color = QColor(200, 255, 200, 100)
            no_color = QColor(25, 200, 200, 100)

        palette = QPalette()
        if flag:
            palette.setColor(QPalette().Base, no_color)
        else:
            palette.setColor(QPalette().Base, yes_color)

        self._directory_edit.setPalette(palette)

    def _on_text_changed(self, text):
        directory = self.get_directory()
        if os.path.exists(directory):
            self.set_error(False)
        else:
            self.set_error(True)

        if not text:
            self._directory_edit.setPalette(lineedit.BaseLineEdit().palette())

        self.directoryChanged.emit(directory)

    def _on_browse(self):
        directory = self.get_directory()
        if not directory:
            placeholder = self._directory_edit.placeholderText()
            if placeholder and placeholder.startswith('example: '):
                example_path = placeholder[0]
                if os.path.exists(example_path):
                    directory = example_path

        filename = qtutils.get_folder(directory, show_files=self._show_files, parent=self)
        filename = path.clean_path(filename)
        if filename and path.is_dir(filename):
            self._directory_edit.setText(filename)
            self.directoryChanged.emit(filename)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains base classes for list views
"""

from __future__ import print_function, division, absolute_import

import os
import glob

from Qt.QtCore import Qt, Signal, QByteArray, QSize, QModelIndex, QPersistentModelIndex, QItemSelectionModel, QMimeData
from Qt.QtCore import QUrl
from Qt.QtWidgets import QListView, QTableView, QTreeView
from Qt.QtGui import QDrag

from tpDcc.libs.python import python
from tpDcc.libs.qt.widgets import models


class BaseListView(QListView, object):
    """
    Base class for list views
    """

    def __ini__(self, parent=None):
        super(BaseListView, self).__init__(parent=parent)


class ListView(BaseListView, object):
    """
    Extends basic QListView functionality
    """

    current_selection_changed = Signal(object, object)
    middle_clicked = Signal(QModelIndex)

    def __init__(self, parent=None):
        super(ListView, self).__init__(parent)

    # region Public Functions
    def appendItem(self, item):
        """
        Adds a new item to the list view model
        :param item: variant
        """

        if self.model():
            self.model().appendItem(item)
    # endregion

    # region Override Functions
    def selectionChanged(self, selected, deselected):
        """
        Override selectionChanged behavior
        :param selected: list<variant>, new selection
        :param deselected: list<variant>, previous selection
        :warning: This method is override to avoid a Qt 4.7.0 QModelIndexList destructor crash.
            See ItemSelection for more information.
        """

        self.current_selection_changed.emit(models.ItemSelection(selected), models.ItemSelection(deselected))
        super(ListView, self).selectionChanged(models.ItemSelection(selected), models.ItemSelection(deselected))

    def selectedIndexes(self):
        """
        Override selectedIndexes behavior
        :warning: This method is override to avoid  a Qt 4.7.0 QModelIndexList destructor crash.
            See ItemSelection for more information.
        """

        item_selection = models.ItemSelection(self.selectionModel().selection())
        return item_selection.indexes()

    def mousePressEvent(self, event):
        """
        Override mousePressEvent behavior to add a custom middle click signal
        :param event: QMouseEvent
        """

        if event.button() == Qt.MidButton:
            model_index = self.indexAt(event.pos())
            if model_index.row() >= 0:
                self.middle_clicked.emit(model_index)
        super(ListView, self).mousePressEvent(event)
    # endregion


class DraggableListView(ListView, object):
    """
    Extends ListView functionality width draggable functionality
    """

    addedItemDir = Signal(list)

    def __init__(self, parent=None):
        super(DraggableListView, self).__init__(parent=parent)

        self._accepted_files = list()

        self.setEditTriggers(QListView.NoEditTriggers)
        self.setDragEnabled(True)
        self.setDragDropMode(QListView.DragOnly)
        self.setDefaultDropAction(Qt.IgnoreAction)
        self.setAcceptDrops(True)
        self.setIconSize(QSize(120, 120))
        self.setViewMode(QListView.IconMode)
        self.setMovement(QListView.Snap)
        self.setResizeMode(QListView.Adjust)
        self.setUniformItemSizes(True)
        self.setMouseTracking(True)

    def startDrag(self, supported_actions):
        """
        Function that is called when we start dragging an item
        This function set ups the mime data and creates a copy of the image
        of the item that is show in cursor position while dragging action
        is enabled
        """

        index = self.currentIndex()
        model = self.model()
        item = model.itemFromIndex(index)
        drag = QDrag(self)
        mime_data = QMimeData()
        mime_data.setData('item_text', QByteArray(str(item.text())))
        mime_data.setData('item_size', QByteArray('%.1f %.1f' % (item.size.width(), item.size.height())))
        url = QUrl.fromLocalFile(item.text())
        mime_data.setUrls(url)
        drag.setMimeData(mime_data)
        pixmap = item.icon().pixmap(50, 50)
        drag.setDragCursor(pixmap, Qt.CopyAction)
        drag.start()

    def dragEnterEvent(self, event):
        mime_data = event.mimeData()
        if mime_data.hasUrls():
            for path in map(lambda x: x.toLocalFile(), mime_data.urls()):
                if os.path.isdir(path) or os.path.isfile(path) and \
                        os.path.splitext(path)[-1].lower() in self._accepted_files:
                    self.setDragDropMode(QListView.InternalMove)
                    event.accept()
                    break
        super(DraggableListView, self).dragEnterEvent(event)

    def dragMoveEvent(self, event):
        event.accept()
        super(DraggableListView, self).dragMoveEvent(event)

    def dropEvent(self, event):
        mime_data = event.mimeData()
        if mime_data.hasUrls():
            send_data = list()
            for path in map(lambda x: x.toLocalFile(), mime_data.urls()):
                for file in self._accepted_files:
                    if os.path.isdir(path) and glob.glob(os.path.join(path, '*' + file))\
                            or os.path.isfile(path) and os.path.splitext(path)[-1].lower() == file:
                        send_data.append(path)
            if send_data:
                self.addedItemDir.emit(send_data)
            self.setDragEnabled(True)
            self.setDragDropMode(QListView.DragOnly)
            self.setDefaultDropAction(Qt.IgnoreAction)
            self.setAcceptDrops(True)
            super(DraggableListView, self).dropEvent(event)

    def set_accepted_files(self, list_files):

        list_files = python.force_list(list_files)

        for f in list_files:
            if not f.startswith('.'):
                f = '.' + f
            self._accepted_files.append(f)


class BaseTableView(QTableView, object):
    def __init__(self, parent=None, **kwargs):
        super(BaseTableView, self).__init__(parent=parent, **kwargs)

        self._last_indexes = list()

        self.installEventFilter(self)

    def focusOutEvent(self, event):
        if self.selectionModel().selectedIndexes():
            for index in self.selectionModel().selectedRows():
                self._last_indexes.append(QPersistentModelIndex(index))
        if self._last_indexes:
            for i in self._last_indexes:
                self.selectionModel().setCurrentIndex(i, QItemSelectionModel.Select)
        event.accept()

    def get_selected_indexes(self):
        """
        Returns the selected indexes
        """

        return self.selectionModel().selectedIndexes()

    def get_selected_rows(self):
        """
        Returns the selected rows
        """

        return self.selectionModel().selectedRows()


class BaseTreeView(QTreeView, object):
    """
    Base class for tree views
    """

    def __init__(self, parent=None):
        super(BaseTreeView, self).__init__(parent=parent)

    def expand_tree(self, root_node):
        """
        Expands all the collapsed elements in a tree starting at the rootNode
        :param root_node: Start node from which we start expanding the tree
        """

        parent = root_node.parent()
        parent_id = self.model().createIndex(parent.row(), 0, parent) if parent else QModelIndex()
        index = self.model().index(root_node.row(), 0, parent_id)
        self.setExpanded(index, True)
        for child in root_node.children:
            self.expand_tree(child)



from functools import partial

import tpDcc
from tpDcc.libs.qt.core import mixin
from tpDcc.libs.qt.widgets import breadcrumb, dividers, message


class BreadcrumbExample(tpDcc.Window, mixin.FieldMixin):
    def __init__(self):
        super(BreadcrumbExample, self).__init__(title='Example for Breadcrumbs')

    def ui(self):
        super(BreadcrumbExample, self).ui()

        message.PopupMessage.config(duration=1)

        entity_list = [
            {'clicked': partial(self._on_show_message, message.PopupMessage.info, 'Go to "Home Page"'),
             'image': 'home'},
            {'text': 'pl', 'clicked': partial(self._on_show_message, message.PopupMessage.info, 'Go to "pl"'),
             'image': 'user'},
            {'text': 'pl_0010', 'clicked': partial(self._on_show_message, message.PopupMessage.info, 'Go to "pl_0010')}
        ]

        no_icon_bread = breadcrumb.BreadcrumbWidget(parent=self)
        no_icon_bread.set_items(entity_list)

        separator_eq = breadcrumb.BreadcrumbWidget(separator='=>', parent=self)
        separator_eq.set_items(entity_list)

        path_bread = breadcrumb.BreadcrumbWidget(separator='/', parent=self)
        path_bread.set_from_path('C:/Users/BreadCrumb/Hello/World')

        frame_bread = breadcrumb.BreadcrumbFrame(parent=self)
        frame_bread.set_items(entity_list)

        self.main_layout.addWidget(dividers.Divider('normal'))
        self.main_layout.addWidget(no_icon_bread)
        self.main_layout.addWidget(dividers.Divider('separator =>'))
        self.main_layout.addWidget(separator_eq)
        self.main_layout.addWidget(dividers.Divider('from path'))
        self.main_layout.addWidget(path_bread)
        self.main_layout.addWidget(dividers.Divider('frame'))
        self.main_layout.addWidget(frame_bread)
        self.main_layout.addStretch()

    def _on_show_message(self, fn, config):
        fn(config, parent=self)


def run():
    return BreadcrumbExample()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains generic functionality when dealing with projects
"""

from __future__ import print_function, division, absolute_import

import os
import logging

from Qt.QtCore import Qt, Signal, QSize
from Qt.QtWidgets import QSizePolicy, QFrame, QPushButton, QMenu, QAction
from Qt.QtGui import QPixmap, QIcon

from tpDcc import dcc
from tpDcc.managers import resources
from tpDcc.core import project as core_project
from tpDcc.core import consts
from tpDcc.libs.python import path, settings, folder, fileio
from tpDcc.libs.qt.core import base, qtutils
from tpDcc.libs.qt.widgets import layouts, search, directory, dividers, buttons, label, tabs, lineedit

LOGGER = logging.getLogger('tpDcc-libs-qt')


def get_project_by_name(projects_path, project_name, project_class=None):
    """
    Returns a project located in the given path and with the given name (if exists)
    :param projects_path: str
    :param project_name: str
    :param project_class: cls
    :return: Project or None
    """

    if not projects_path or not os.path.isdir(projects_path):
        LOGGER.warning('Projects Path "{}" does not exist!'.format(projects_path))
        return None

    all_projects = get_projects(projects_path, project_class=project_class)
    for project in all_projects:
        if project.name == project_name:
            return project

    return None


def get_projects(projects_path, project_class=None):
    """
    Returns all projects located in given path
    :param projects_path: str
    :param project_class: cls
    :return: list(Project)
    """

    if not project_class:
        project_class = Project

    projects_found = list()

    if not projects_path or not os.path.isdir(projects_path):
        LOGGER.warning('Projects Path {} is not valid!'.format(projects_path))
        return projects_found

    for root, dirs, files in os.walk(projects_path):
        if consts.PROJECTS_NAME in files:
            new_project = project_class.create_project_from_data(
                path.join_path(root, consts.PROJECTS_NAME))
            if new_project is not None:
                projects_found.append(new_project)

    return projects_found


class Project(base.BaseWidget):
    projectOpened = Signal(object)
    projectRemoved = Signal(str)
    projectImageChanged = Signal(str)

    def __init__(self, project_data, parent=None):

        self._project_data = project_data

        super(Project, self).__init__(parent)

    # ============================================================================================================
    # PROPERTIES
    # ============================================================================================================

    @property
    def name(self):
        return self._project_data.name

    @property
    def path(self):
        return self._project_data.path

    @property
    def full_path(self):
        return self._project_data.full_path

    @property
    def settings(self):
        return self._project_data.settings

    @property
    def project_data(self):
        return self._project_data

    # ============================================================================================================
    # CLASS FUNCTIONS
    # ============================================================================================================

    @classmethod
    def create_project_from_data(cls, project_data_path):
        """
        Creates a new project using a project data JSON file
        :param project_data_path: str, path where project JSON data file is located
        :return: Project
        """

        if project_data_path is None or not path.is_file(project_data_path):
            LOGGER.warning('Project Data Path {} is not valid!'.format(project_data_path))
            return None

        project_data = settings.JSONSettings()
        project_options = settings.JSONSettings()
        project_dir = path.get_dirname(project_data_path)
        project_name = path.get_basename(project_data_path)
        project_data.set_directory(project_dir, project_name)
        project_options.set_directory(project_dir, 'options.json')

        project_name = project_data.get('name')
        project_path = path.get_dirname(path.get_dirname(project_data_path))
        project_image = project_data.get('image')

        LOGGER.debug('New Project found [{}]: {}'.format(project_name, project_path))
        project_data = core_project.ProjectData(
            name=project_name, project_path=project_path, settings=project_data, options=project_options)

        new_project = cls(project_data=project_data)
        if project_image:
            new_project.set_image(project_image)

        return new_project

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def ui(self):
        super(Project, self).ui()

        self.setMaximumWidth(qtutils.dpi_scale(160))
        self.setMaximumHeight(qtutils.dpi_scale(200))

        widget_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        main_frame = QFrame()
        main_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
        main_frame.setLineWidth(1)
        main_frame.setLayout(widget_layout)
        self.main_layout.addWidget(main_frame)

        self.project_btn = QPushButton('', self)
        self.project_btn.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.project_btn.setIconSize(QSize(120, 120))
        project_lbl = label.BaseLabel(self.name, parent=self)
        project_lbl.setObjectName('projectLabel')
        project_lbl.setAlignment(Qt.AlignCenter)
        widget_layout.addWidget(self.project_btn)
        widget_layout.addWidget(project_lbl)

    def setup_signals(self):
        self.project_btn.clicked.connect(self._on_open_project)

    def contextMenuEvent(self, event):
        menu = QMenu(self)
        remove_icon = resources.icon(name='delete')
        remove_action = QAction(remove_icon, 'Remove', menu)
        remove_tooltip = 'Delete selected project'
        remove_action.setStatusTip(remove_tooltip)
        remove_action.setToolTip(remove_tooltip)
        remove_action.triggered.connect(self._on_remove_project)

        folder_icon = resources.icon(name='open_folder', extension='png')
        folder_action = QAction(folder_icon, 'Open in Browser', menu)
        open_project_in_explorer_tooltip = 'Open project folder in explorer'
        folder_action.setStatusTip(open_project_in_explorer_tooltip)
        folder_action.setToolTip(open_project_in_explorer_tooltip)
        folder_action.triggered.connect(self._on_open_in_browser)

        image_icon = resources.icon(name='picture', extension='png')
        set_image_action = QAction(image_icon, 'Set Project Image', menu)
        set_project_image_tooltip = 'Set the image used by the project'
        set_image_action.setToolTip(set_project_image_tooltip)
        set_image_action.setStatusTip(set_project_image_tooltip)
        set_image_action.triggered.connect(self._on_set_project_image)

        for action in [remove_action, None, folder_action, None, set_image_action]:
            if action is None:
                menu.addSeparator()
            else:
                menu.addAction(action)

        menu.exec_(self.mapToGlobal(event.pos()))

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def open(self):
        """
        Opens project
        """

        self._on_open_project()

    def has_option(self, name, group=None):
        """
        Returns whether the current object has given option or not
        :param name: str, name of the option
        :param group: variant, str || None, group of the option (optional)
        :return: bool
        """

        if not self._project_data:
            return False

        return self._project_data.has_option(name=name, group=group)

    def add_option(self, name, value, group=None, option_type=None):
        """
        Adds a new option to the options file
        :param name: str, name of the option
        :param value: variant, value of the option
        :param group: variant, str || None, group of the option (optional)
        :param option_type: variant, str || None, option type (optional)
        """

        if not self._project_data:
            return

        self._project_data.add_option(name, value, group=group, option_type=option_type)

    def get_option(self, name, group=None, default=None):
        """
        Returns option by name and group
        :param name: str, name of the option we want to retrieve
        :param group: variant, str || None, group of the option (optional)
        :return: variant
        """

        if not self._project_data:
            return

        return self._project_data.get_option(name, group=group, default=default)

    def reload_options(self):
        """
        Reload settings
        """

        if not self._project_data:
            return

        self._project_data.reload_options()

    def clear_options(self):
        """
        Clears all the options
        """

        if not self._project_data:
            return

        self._project_data.clear_options()

    def set_image(self, encoded_image):

        from tpDcc.libs.qt.core import image

        if not encoded_image:
            return

        encoded_image = encoded_image.encode('utf-8')
        project_icon = QIcon(QPixmap.fromImage(image.base64_to_image(encoded_image)))
        if project_icon.isNull():
            project_icon = resources.icon('tpDcc')
        self.project_btn.setIcon(project_icon)

    def remove(self, force=False):
        if not path.is_dir(self.full_path):
            LOGGER.warning('Impossible to remove Project Path: {}'.format(self.full_path))
            return False

        project_name = self.project_data.name
        project_path = self.project_data.path

        if not force:
            result = qtutils.get_permission(
                message='Are you sure you want to delete project: "{}"'.format(self.name),
                title='Deleting Project', cancel=False, parent=self)
            if not result:
                return

        valid_delete = folder.delete_folder(folder_name=project_name, directory=project_path)
        if valid_delete is None:
            return False

        return True

    def load_project_data(self):
        """
        Return dictionary data contained in the project
        :return: dict
        """

        if not self.settings:
            return

        return self.settings.data()

    def get_project_nodes(self):
        """
        Returns path where nodes should be stored
        :return: str
        """

        return [os.path.join(self.full_path, 'nodes'), os.path.join(self.full_path, 'components')]

    def get_options(self):
        """
        Returns all options contained in the project
        :return: str
        """

        return self._project_data.get_options()

    def get_project_image(self):
        """
        Returns the image used by the project
        :return: QPixmap
        """

        return self._project_data.get_project_image()

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_open_project(self):
        """
        Internal callback function that is called when a project is opened
        """

        LOGGER.info('Loading project "{}" ...'.format(self.full_path))
        self.projectOpened.emit(self)

    def _on_remove_project(self):
        """
        Internal callback function that is called when a project is removed
        """

        valid_remove = self.remove()
        if valid_remove:
            self.projectRemoved.emit(self.name)

    def _on_open_in_browser(self):
        """
        Internal callback function that is called when a project is browsed
        """

        fileio.open_browser(self.full_path)

    def _on_set_project_image(self):
        """
        Internal callback function that is called when project image is set
        """

        image_file = dcc.select_file_dialog(
            title='Select Project Image File',
            pattern="PNG Files (*.png)")

        if image_file is None or not path.is_file(image_file):
            LOGGER.warning('Selected Image "{}" is not valid!'.format(image_file))
            return

        valid_change = self._project_data.set_project_image(image_file)

        if valid_change:
            project_image = self._project_data.settings.get('image')
            if project_image:
                self.set_image(project_image)
            self.projectImageChanged.emit(image_file)


class TemplateData(object):

    PROJECT_CLASS = Project

    def __init__(self, name='New Template'):
        self._name = name

    def get_name(self):
        return self._name

    name = property(get_name)

    def create_project(self, project_name, project_path):
        if not self.PROJECT_CLASS:
            LOGGER.warning('Impossible to create because project class is not defined!')
            return None

        project_data = settings.JSONSettings()
        project_options = settings.JSONSettings()
        LOGGER.debug('New Project found [{}]: {}'.format(project_name, project_path))
        project_data = core_project.ProjectData(
            name=project_name, project_path=project_path, settings=project_data, options=project_options)
        project_data.create_project()

        new_project = self.PROJECT_CLASS(project_data=project_data)

        return new_project


class Template(base.BaseWidget):
    templateChecked = Signal(object)

    def __init__(self, parent=None):
        super(Template, self).__init__(parent=parent)

    def ui(self):
        super(Template, self).ui()

        self.setMaximumWidth(160)
        self.setMaximumHeight(200)

        widget_layout = layouts.VerticalLayout(spacing=0, margins=(2, 2, 2, 2))
        main_frame = QFrame()
        main_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
        main_frame.setLineWidth(1)
        main_frame.setLayout(widget_layout)
        self.main_layout.addWidget(main_frame)

        self.template_btn = QPushButton('', self)
        self.template_btn.setCheckable(True)
        self.template_btn.setIcon(self.get_icon())
        self.template_btn.setIconSize(QSize(120, 120))
        template_lbl = label.BaseLabel(self.name, parent=self)
        template_lbl.setObjectName('templateLabel')
        template_lbl.setAlignment(Qt.AlignCenter)
        widget_layout.addWidget(self.template_btn)
        widget_layout.addWidget(template_lbl)

    def setup_signals(self):
        self.template_btn.toggled.connect(self._on_selected_template)

    def get_icon(self):
        return resources.icon(name='project', extension='png')

    def _on_selected_template(self, template):
        self.templateChecked.emit(self)


class BlankTemplateData(TemplateData, object):
    def __init__(self, name='Blank'):
        super(BlankTemplateData, self).__init__(name=name)

    def get_name(self):
        return self._name

    name = property(get_name)

    def create_project(self, project_name, project_path):
        new_project = super(
            BlankTemplateData, self).create_project(project_name=project_name, project_path=project_path)
        return new_project


class BlankTemplate(Template, BlankTemplateData):
    def __init__(self, parent=None):
        BlankTemplateData.__init__(self)
        Template.__init__(self, parent=parent)


class TemplatesViewer(base.BaseWidget, object):

    STANDARD_TEMPLATES = [BlankTemplate]
    selectedTemplate = Signal(object)

    def __init__(self, project_class, parent=None):
        self._project_class = project_class
        super(TemplatesViewer, self).__init__(parent=parent)

        self._init_standard_templates()

    def get_main_layout(self):
        main_layout = layouts.FlowLayout(parent=self)
        return main_layout

    def get_widgets(self):
        all_widgets = list()

        for i in range(self.main_layout.count()):
            widget_item = self.main_layout.itemAt(i)
            if not widget_item:
                continue
            w = widget_item.widget()
            all_widgets.append(w)

        return all_widgets

    def add_template(self, template_widget):
        if template_widget is None:
            return

        template_widget.PROJECT_CLASS = self._project_class
        template_widget.templateChecked.connect(self._on_template_selected)

        self.main_layout.addWidget(template_widget)

    def clear_templates(self):
        qtutils.clear_layout(self.main_layout)

    def _init_standard_templates(self):
        for template in self.STANDARD_TEMPLATES:
            new_template = template()
            self.add_template(new_template)

    def _on_template_selected(self, template):
        self.selectedTemplate.emit(template)


class OpenProjectWidget(base.BaseWidget, object):
    projectOpened = Signal(object)
    projectsPathChanged = Signal(str)

    def __init__(self, project_class, projects_path=None, parent=None):

        self._project_class = project_class
        self._projects_path = projects_path

        super(OpenProjectWidget, self).__init__(parent=parent)

        self._update_ui()

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def get_main_layout(self):
        return layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))

    def ui(self):
        super(OpenProjectWidget, self).ui()

        self._search_widget = search.SearchFindWidget()
        self._search_widget.set_placeholder_text('Filter Projects ...')

        self._projects_list = ProjectViewer(project_class=self._project_class)
        self._projects_list.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        buttons_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        buttons_layout.setAlignment(Qt.AlignCenter)

        buttons_layout1 = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        buttons_layout1.setAlignment(Qt.AlignLeft)
        self._browse_widget = directory.SelectFolder(label_text='Projects Path', use_app_browser=True, use_icon=True)
        buttons_layout1.addWidget(self._browse_widget)

        buttons_layout.addLayout(buttons_layout1)

        self.main_layout.addWidget(self._search_widget)
        self.main_layout.addWidget(self._projects_list)
        self.main_layout.addLayout(buttons_layout)

    def setup_signals(self):
        self._search_widget.textChanged.connect(self._on_search_project)
        self._browse_widget.directoryChanged.connect(self._on_directory_browsed)
        self._projects_list.projectOpened.connect(self.projectOpened.emit)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def get_projects_list(self):
        """
        Returns projects list widget
        :return: ProjectViewer
        """

        return self._projects_list

    def set_projects_path(self, projects_path):
        """
        Sets the path where we want to search for projects
        :param projects_path: str
        """

        self._projects_path = projects_path
        # We set the projects path of the projects list after updating UI
        self._projects_list.set_projects_path(self._projects_path)

        self._update_ui()

        self.projectsPathChanged.emit(self._projects_path)

    def update_projects(self):
        """
        Updates all available projects
        """

        self._projects_list.update_projects()

    def add_project(self, project_name):
        self._projects_list.add_project(project_name)

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _update_ui(self):
        """
        Internal function that updates UI
        """

        if not self._projects_path or not os.path.isdir(self._projects_path):
            return

        self._browse_widget.set_directory(directory=self._projects_path)
        self.update_projects()

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_search_project(self, project_text):
        """
        Internal callback function that is called when the user types in the search projects filter
        :param project_text: str
        """

        for project in self._projects_list.get_widgets():
            project.setVisible(project_text.lower() in project.name.lower())

    def _on_directory_browsed(self, projects_path):
        """
        Internal callback function that is triggered when the user browses a new projects path
        :param projects_path: str
        """

        if not projects_path or not path.is_dir(projects_path):
            return

        self.set_projects_path(projects_path)
        self._update_ui()


class NewProjectWidget(base.BaseWidget, object):

    TEMPLATES_VIEWER_CLASS = TemplatesViewer

    projectCreated = Signal(str)

    def __init__(self, project_class, projects_path=None, parent=None):

        self._selected_template = None
        self._project_class = project_class
        self._projects_path = projects_path

        super(NewProjectWidget, self).__init__(parent=parent)

    def get_main_layout(self):
        return layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))

    def ui(self):
        super(NewProjectWidget, self).ui()

        self._search_widget = search.SearchFindWidget()
        self._search_widget.set_placeholder_text('Filter Templates ...')

        self._templates_list = self.TEMPLATES_VIEWER_CLASS(project_class=self._project_class)
        self._templates_list.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        project_layout = layouts.HorizontalLayout(spacing=1, margins=(0, 0, 0, 0))

        project_line_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        project_layout.addLayout(project_line_layout)
        self._project_line = lineedit.BaseLineEdit(parent=self)
        self._project_line.setPlaceholderText('Project Path')
        self._project_line.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self._project_btn = directory.SelectFolderButton(text='', use_app_browser=True)
        project_line_layout.addWidget(self._project_line)
        project_line_layout.addWidget(self._project_btn)
        self._name_line = lineedit.BaseLineEdit(parent=self)
        self._name_line.setPlaceholderText('Project Name')
        project_line_layout.addWidget(dividers.get_horizontal_separator_widget())
        project_line_layout.addWidget(self._name_line)
        self._create_btn = buttons.BaseButton('Create', parent=self)
        self._create_btn.setIcon(resources.icon('create'))
        project_line_layout.addSpacing(10)
        project_line_layout.addWidget(self._create_btn)

        self.main_layout.addWidget(self._search_widget)
        self.main_layout.addWidget(self._templates_list)
        self.main_layout.addLayout(project_layout)

    def setup_signals(self):
        self._search_widget.textChanged.connect(self._on_search_template)
        self._templates_list.selectedTemplate.connect(self._on_selected_template)
        self._project_btn.directoryChanged.connect(self._on_directory_browsed)
        self._create_btn.clicked.connect(self._on_create)

    @property
    def templates_list(self):
        return self._templates_list

    def set_projects_path(self, projects_path):
        """
        Set the path where projects are located
        """

        self._projects_path = projects_path
        self._update_ui()

    def _update_ui(self):
        """
        Update UI based on the stored settings if exists
        """

        if not self._projects_path or not os.path.isdir(self._projects_path):
            return

        self._project_line.setText(self._projects_path)
        self._project_btn.init_directory = self._projects_path

    def _on_search_template(self, template_text):
        for template in self._templates_list.get_widgets():
            template.setVisible(template_text.lower() in template.name.lower())

    def _on_selected_template(self, template):
        self._selected_template = template

    def _on_directory_browsed(self, dir):
        if not dir or not path.is_dir(dir):
            return

        self._project_line.setText(str(dir))

    def _on_create(self):
        project_path = self._project_line.text()
        project_name = self._name_line.text()
        if not project_path or not path.is_dir(project_path) or not project_name:
            LOGGER.warning('Project Path: {} or Project Name: {} are not valid!'.format(project_path, project_name))
            return
        if self._selected_template is None:
            LOGGER.warning('No Template selected, please select one first ...')
            return

        new_project = self._selected_template.create_project(project_name=project_name, project_path=project_path)
        if new_project is not None:
            LOGGER.debug(
                'Project {} created successfully on path {}'.format(new_project.name, new_project.path))
            self._name_line.setText('')
            self.projectCreated.emit(new_project.name)
            return new_project

        return None


class ProjectViewer(base.BaseWidget, object):
    projectOpened = Signal(object)

    def __init__(self, project_class, projects_path=None, parent=None):
        self._project_class = project_class
        self._projects_path = None
        super(ProjectViewer, self).__init__(parent=parent)

        self.set_projects_path(projects_path)

    def get_main_layout(self):
        main_layout = layouts.FlowLayout(parent=self, spacing_x=0, spacing_y=0)
        return main_layout

    def set_projects_path(self, projects_path):
        """
        Set the path where projects are located
        :param projects_path: str
        """

        self._projects_path = projects_path
        self.update_projects()

    def add_project_widget(self, project_widget):
        if project_widget is None:
            return

        project_widget.projectOpened.connect(self.projectOpened.emit)
        project_widget.projectRemoved.connect(self._on_project_removed)

        self.main_layout.addWidget(project_widget)

    def get_widgets(self):
        all_widgets = list()

        for i in range(self.main_layout.count()):
            widget_item = self.main_layout.itemAt(i)
            if not widget_item:
                continue
            w = widget_item.widget()
            all_widgets.append(w)

        return all_widgets

    def get_project_by_name(self, project_name):
        for w in self.get_widgets():
            if w.name == project_name:
                return w

        return None

    def update_projects(self):

        qtutils.clear_layout(self.main_layout)

        if not self._projects_path or not os.path.isdir(self._projects_path):
            return

        projects_found = get_projects(self._projects_path, project_class=self._project_class)
        for project_found in projects_found:
            self.add_project_widget(project_found)

    def add_project(self, project_name):
        if not self._projects_path or not os.path.isdir(self._projects_path):
            return

        projects_found = get_projects(self._projects_path, project_class=self._project_class)
        for project_found in projects_found:
            if project_found.name == project_name:
                self.add_project_widget(project_found)
                break

    def _on_project_removed(self, project_name):
        project_widget = self.get_project_by_name(project_name)
        if not project_widget:
            return

        project_widget.setParent(None)
        project_widget.deleteLater()


class ProjectWidget(base.BaseWidget, object):

    PROJECT_CLASS = Project
    OPEN_PROJECT_CLASS = OpenProjectWidget
    NEW_PROJECT_CLASS = NewProjectWidget

    projectOpened = Signal(object)
    projectsPathChanged = Signal(str)

    def __init__(self, projects_path=None, parent=None):

        self._projects_path = None
        self._history = None

        super(ProjectWidget, self).__init__(parent=parent)

        self.set_projects_path(projects_path)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def get_main_layout(self):
        return layouts.VerticalLayout(spacing=2, margins=(2, 2, 2, 2))

    def ui(self):
        super(ProjectWidget, self).ui()

        self._tab_widget = tabs.BaseTabWidget(parent=self)
        self._open_project = self.OPEN_PROJECT_CLASS(project_class=self.PROJECT_CLASS)
        self._new_project = self.NEW_PROJECT_CLASS(project_class=self.PROJECT_CLASS)
        self._tab_widget.addTab(self._open_project, 'Projects')
        self._tab_widget.addTab(self._new_project, 'New Project')
        self.main_layout.addWidget(self._tab_widget)

    def setup_signals(self):
        self._open_project.projectOpened.connect(self.projectOpened.emit)
        self._open_project.projectsPathChanged.connect(self.projectsPathChanged.emit)
        self._new_project.projectCreated.connect(self._on_project_created)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def set_projects_path(self, projects_path):
        """
        Sets the path where we want to search for projects
        :param projects_path: str
        """

        self._projects_path = projects_path
        self._new_project.set_projects_path(projects_path)
        self._open_project.set_projects_path(projects_path)

    def get_project_by_name(self, project_name, force_update=True):
        """
        Returns project by its name
        :param project_name: str
        :param force_update: bool
        :return: Project
        """

        if force_update:
            self._open_project.get_projects_list().update_projects()

        projects_list = self._open_project.get_projects_list()
        return projects_list.get_project_by_name(project_name)

    def open_project(self, project_name):
        """
        Opens project with given name
        :param project_name: str
        :return: Project
        """

        project_found = self.get_project_by_name(project_name)
        if project_found:
            project_found.open()
            return project_found

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_project_created(self, new_project):
        self._tab_widget.setCurrentIndex(0)
        self._open_project.add_project(new_project)



from functools import partial

import tpDcc
from tpDcc.libs.qt.core import mixin
from tpDcc.libs.qt.widgets import tabs, label, message, badge, buttons


class MenuTabWidgetExample(tpDcc.Window, mixin.FieldMixin):
    def __init__(self):
        super(MenuTabWidgetExample, self).__init__(title='Example for Menu Tab Widgets')

    def ui(self):
        super(MenuTabWidgetExample, self).ui()

        item_list = [
            {'text': 'Overview', 'image': 'home', 'clicked': partial(message.PopupMessage.info, 'A', parent=self)},
            {'text': 'B', 'image': 'user', 'clicked': partial(message.PopupMessage.info, 'B', parent=self)},
            {'text': 'Notice', 'image': 'notification', 'clicked': partial(message.PopupMessage.info, 'C', parent=self)}
        ]

        tool_bar = tabs.MenuTabWidget()
        tool_bar.insert_widget(label.BaseLabel('Hello World!').h4().secondary().strong())
        tool_bar.append_widget(
            badge.Badge.create_dot(
                show=True, widget=buttons.BaseToolButton().image('user').icon_only().large(), parent=self))
        self._content_widget = label.BaseLabel()
        for index, data_dict in enumerate(item_list):
            tool_bar.add_menu(data_dict, index)
        tool_bar.tool_btn_grp.checked = False

        self.main_layout.addWidget(tool_bar)
        self.main_layout.addWidget(self._content_widget)


def run():
    return MenuTabWidgetExample()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains widget used to handle theme related preferences (PreferencesWidget)
"""

from __future__ import print_function, division, absolute_import

from functools import partial

from Qt.QtCore import Qt
from Qt.QtWidgets import QWidget, QScrollArea
from Qt.QtGui import QColor, QPixmap, QIcon, QPainter

from tpDcc.managers import resources
from tpDcc.libs.resources.core import color
from tpDcc.libs.qt.core import preferences
from tpDcc.libs.qt.widgets import layouts, accordion, combobox, label, color as qt_color


class ThemePreferenceWidget(preferences.CategoryWidgetBase, object):
    """
    Widget used to handle theme related preferences (PreferencesWidget)
    """

    CATEGORY = 'Theme'

    def __init__(self, theme, parent=None):

        self._theme = theme
        self._original_options = self._theme.options().copy()
        self._color_buttons = list()

        super(ThemePreferenceWidget, self).__init__(parent=parent)

    def ui(self):
        super(ThemePreferenceWidget, self).ui()

        theme_accordion = accordion.AccordionWidget(parent=self)
        theme_accordion.rollout_style = accordion.AccordionStyle.SQUARE
        self.main_layout.addWidget(theme_accordion)

        theme_accordion.add_item('General', self._setup_general_tab(), icon=resources.icon('settings'))
        theme_accordion.add_item('Colors', self._setup_colors_tab(), True, icon=resources.icon('palette'))
        theme_accordion.add_item('Fonts', self._setup_fonts_tab(), True, icon=resources.icon('font_size'))

    def _setup_general_tab(self):
        general_widget = QWidget()
        general_layout = layouts.VerticalLayout(spacing=2, margins=(2, 2, 2, 2))
        general_widget.setLayout(general_layout)

        self._themes_combobox = combobox.BaseComboBox(parent=self)
        all_themes = resources.get_all_resources_of_type(resources.ResourceTypes.THEME)

        for i, theme in enumerate(all_themes):
            accent_color_hex = theme.accent_color
            accent_color = color.Color.hex_to_qcolor(
                accent_color_hex[1:] if accent_color_hex.startswith('#') else accent_color_hex)
            background_color_hex = theme.background_color
            background_color = color.Color.hex_to_qcolor(
                background_color_hex[1:] if accent_color_hex.startswith('#') else background_color_hex)
            accent_color_pixmap = QPixmap(25, 25)
            background_color_pixmap = QPixmap(25, 25)
            accent_color_pixmap.fill(accent_color)
            background_color_pixmap.fill(background_color)
            color_pixmap = QPixmap(50, 25)
            painter = QPainter(color_pixmap)
            painter.drawPixmap(0, 0, 25, 25, accent_color_pixmap)
            painter.drawPixmap(25, 0, 25, 25, background_color_pixmap)
            painter.end()

            color_icon = QIcon(color_pixmap)
            self._themes_combobox.addItem(color_icon, theme.name())

        general_layout.addWidget(self._themes_combobox)
        general_layout.addStretch()

        return general_widget

    def setup_signals(self):
        self._themes_combobox.currentIndexChanged.connect(self._on_current_theme_changed)

    def show_widget(self, settings):
        if not settings or not self._theme:
            return

        theme_settings = dict()
        attributes_to_load = self._theme.get_color_attribute_names()

        theme_name = settings.get('name', setting_group=self.CATEGORY)
        for attr_name in attributes_to_load:
            value = settings.get(attr_name, setting_group=self.CATEGORY)
            if not value:
                continue
            theme_settings[attr_name] = value

        if theme_name:
            self._theme.set_name(theme_name)
        self._theme.set_settings(theme_settings)

        self._update_color_buttons_color()

    def serialize(self, settings):
        if not settings or not self._theme:
            return

        theme_options = self._theme.options()
        attributes_to_save = self._theme.get_color_attribute_names()
        theme_name = self.theme().name()
        settings.set('name', theme_name)
        for attribute_to_save in attributes_to_save:
            if attribute_to_save not in theme_options:
                continue
            settings.set(attribute_to_save, theme_options[attribute_to_save])
        # settings.endGroup()

        return theme_options

    def reset(self):
        self._theme.set_settings(self._original_options)
        self._update_color_buttons_color()
        theme_name = self._theme.name()
        theme_index = self._themes_combobox.findText(theme_name)
        if theme_index == -1:
            return

        self._themes_combobox.blockSignals(True)
        try:
            self._themes_combobox.setCurrentIndex(theme_index)
        finally:
            self._themes_combobox.blockSignals(False)
        self._themes_combobox.repaint()
        self._themes_combobox.update()

    def _setup_colors_tab(self):
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        colors_widget = QWidget()
        colors_layout = layouts.GridLayout(spacing=2, margins=(2, 2, 2, 2))
        colors_layout.setAlignment(Qt.AlignTop)
        colors_widget.setLayout(colors_layout)
        scroll_area.setWidget(colors_widget)

        color_attribute_names = self._theme.get_color_attribute_names() or list()
        for i, color_attribute_name in enumerate(color_attribute_names):
            if not hasattr(self._theme, color_attribute_name):
                continue
            label, selector = self._add_color_widget(color_attribute_name, getattr(self._theme, color_attribute_name))
            colors_layout.addWidget(label, i, 0, Qt.AlignRight)
            colors_layout.addWidget(selector, i, 1)

        return scroll_area

    def _setup_fonts_tab(self):
        fonts_widget = QWidget()
        fonts_layout = layouts.VerticalLayout(spacing=2, margins=(2, 2, 2, 2))
        fonts_widget.setLayout(fonts_layout)

        return fonts_widget

    def _add_color_widget(self, color_name, color_value):
        color_label = label.BaseLabel('{}: '.format(color_name.replace('_', ' ').title()), parent=self)
        color_button = qt_color.ColorSelector(parent=self)
        color_button.setProperty('color_attribute', color_name)
        # color_button.setProperty('tooltip_help', {'title': 'Color', 'description': 'This is a color'})
        color_button.set_color(QColor(color_value))
        color_button.colorChanged.connect(partial(self._on_color_changed, color_name))
        self._color_buttons.append(color_button)

        return color_label, color_button

    def _update_color_buttons_color(self):
        if not self._color_buttons or not self._theme:
            return

        theme_name = self._theme.name()
        theme_index = self._themes_combobox.findText(theme_name)
        if theme_index > -1:
            self._themes_combobox.blockSignals(True)
            try:
                self._themes_combobox.setCurrentIndex(theme_index)
            finally:
                self._themes_combobox.blockSignals(False)

        for color_button in self._color_buttons:
            color_attribute = color_button.property('color_attribute')
            if not color_attribute:
                continue
            if hasattr(self._theme, color_attribute):
                color_button.blockSignals(True)
                try:
                    color_button.set_color(QColor(getattr(self._theme, color_attribute)))
                finally:
                    color_button.blockSignals(False)

    def _on_color_changed(self, property_name, new_color):
        if not self._theme:
            return

        hex_color = new_color.name()

        if hasattr(self._theme, property_name):
            setattr(self._theme, property_name, hex_color)
            self._theme.update()

    def _on_current_theme_changed(self, theme_index):

        theme_name = self._themes_combobox.itemText(theme_index)
        all_themes = resources.get_all_resources_of_type(resources.ResourceTypes.THEME)
        for theme_found in all_themes:
            if not theme_found or not theme_found.name() == theme_name:
                continue
            theme_options = theme_found.options()
            self._theme.set_name(theme_name)
            self._theme.set_settings(theme_options)
            self._update_color_buttons_color()
            break

        return True

    #                 "name": "accent_color",
    #                 "type": "color",
    #                 "value": accent_color,
    #                 "colors": [
    #                     "rgb(230, 80, 80, 255)",
    #                     "rgb(230, 125, 100, 255)",
    #                     "rgb(230, 120, 40)",
    #                     "rgb(240, 180, 0, 255)",
    #                     "rgb(80, 200, 140, 255)",
    #                     "rgb(50, 180, 240, 255)",
    #                     "rgb(110, 110, 240, 255)",
    #                 ]
    #                 "name": "background_color",
    #                 "type": "color",
    #                 "value": background_color,
    #                 "colors": [
    #                     "rgb(40, 40, 40)",
    #                     "rgb(68, 68, 68)",
    #                     "rgb(80, 60, 80)",
    #                     "rgb(85, 60, 60)",
    #                     "rgb(60, 75, 75)",
    #                     "rgb(60, 64, 79)",
    #                     "rgb(245, 245, 255)",
    #                 ]



from Qt.QtCore import Qt
from Qt.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QLabel, QPushButton, QCheckBox, QDockWidget, QGroupBox


class BaseWidget(QWidget):
    def __init__(self, *args, **kwargs):
        super(BaseWidget, self).__init__(*args, **kwargs)

        self.main_layout = QVBoxLayout()
        self.setLayout(self.main_layout)

        # Label
        label_tooltip = QLabel('I have a tooltip', self)
        label_tooltip.setToolTip('Hello ToolTip!')

        # Button
        normal_button = QPushButton('Normal', self)
        disabled_button = QPushButton('Disabled', self)
        disabled_button.setEnabled(False)

        # CheckBox
        checkboxes_layout = QHBoxLayout()
        checkbox_on = QCheckBox('On', self)
        checkbox_on.setChecked(True)
        checkbox_off = QCheckBox('Off', self)
        checkbox_on_disabled = QCheckBox('On', self)
        checkbox_on_disabled.setChecked(True)
        checkbox_on_disabled.setEnabled(False)
        checkbox_off_disabled = QCheckBox('Off', self)
        checkbox_off_disabled.setEnabled(False)
        checkbox_tri_state = QCheckBox('TriState', self)
        checkbox_tri_state.setTristate(True)
        checkboxes_layout.addWidget(checkbox_on)
        checkboxes_layout.addWidget(checkbox_off)
        checkboxes_layout.addWidget(checkbox_on_disabled)
        checkboxes_layout.addWidget(checkbox_off_disabled)
        checkboxes_layout.addWidget(checkbox_tri_state)

        # GroupBox
        group_box = QGroupBox(self)
        group_box_layout = QVBoxLayout()
        group_box.setLayout(group_box_layout)
        group_button = QPushButton('GrupBox Button')
        group_box_layout.addWidget(group_button)

        # Dock Widget
        dock_label = QLabel('Hello Dock!', self)
        dock_widget = QDockWidget(self)
        dock_widget.setWidget(dock_label)
        self.parent().addDockWidget(Qt.BottomDockWidgetArea, dock_widget)

        self.main_layout.addWidget(label_tooltip)
        self.main_layout.addWidget(normal_button)
        self.main_layout.addWidget(disabled_button)
        self.main_layout.addLayout(checkboxes_layout)
        self.main_layout.addWidget(group_box)
        self.main_layout.addStretch()


if __name__ == '__main__':

    from Qt.QtWidgets import QMainWindow

    import tpDcc.loader
    from tpDcc.managers import resources
    from tpDcc.libs.qt.core import contexts

    tpDcc.loader.init()

    with contexts.application():
        default_theme = resources.theme('default')
        window = QMainWindow()
        widget = BaseWidget(window)
        window.setCentralWidget(widget)
        window.statusBar().showMessage('Status Bar')
        window.setStyleSheet(default_theme.stylesheet())
        window.show()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes to create different kind of labels
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Signal, Property, QRect, QSize, QPropertyAnimation
from Qt.QtWidgets import QSizePolicy, QLabel, QLineEdit, QStyleOption
from Qt.QtGui import QFontMetrics, QTextCursor, QTextDocument, QPainter

from tpDcc.libs.qt.core import qtutils
from tpDcc.libs.qt.widgets import graphicseffects


class BaseLabel(QLabel, object):

    clicked = Signal()

    class Levels(object):
        H1 = 1
        H2 = 2
        H3 = 3
        H4 = 4

    class Types(object):
        SECONDARY = 'secondary'
        WARNING = 'warning'
        DANGER = 'danger'

    def __init__(self, text='', parent=None):
        super(BaseLabel, self).__init__(text, parent)

        self._actual_text = text
        self._underline = False
        self._mark = False
        self._delete = False
        self._strong = False
        self._code = False
        self._elide_mode = Qt.ElideNone

        self._type = ''
        self._level = 0

        self.setTextInteractionFlags(Qt.TextBrowserInteraction | Qt.LinksAccessibleByMouse)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        self.setProperty('label_text', text)

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    def _get_type(self):
        """
        Returns label type
        :return: str, BaseLabel.Types
        """

        return self._type

    def _set_type(self, value):
        """
        Sets label type
        :param value: BaseLabel.Types
        """

        self._type = value
        self.style().polish(self)

    def _get_level(self):
        """
        Returns label level
        :return: str
        """

        return self._level

    def _set_level(self, value):
        """
        Sets label level
        :param value: itn (BaseLabel.Levels)
        """

        self._level = value
        self.style().polish(self)

    def _get_underline(self):
        """
        Returns whether or not label is using an underline style
        :return: bool
        """

        return self._underline

    def _set_underline(self, flag):
        """
         Sets label to use an underline style
         :return: bool
         """

        self._underline = flag
        self.style().polish(self)

    def _get_delete(self):
        """
        Returns whether or not label is using a delete style
        :return: bool
        """

        return self._delete

    def _set_delete(self, flag):
        """
         Sets label to use a delete style
         :return: bool
         """

        self._delete = flag
        self.style().polish(self)

    def _get_strong(self):
        """
        Returns whether or not label is using a strong style
        :return: bool
        """

        return self._strong

    def _set_strong(self, flag):
        """
         Sets label to use a strong style
         :return: bool
         """

        self._strong = flag
        self.style().polish(self)

    def _get_mark(self):
        """
        Returns whether or not label is using a mark style
        :return: bool
        """

        return self._mark

    def _set_mark(self, flag):
        """
         Sets label to use a mark style
         :return: bool
         """

        self._mark = flag
        self.style().polish(self)

    def _get_code(self):
        """
        Returns whether or not label is using a code style
        :return: bool
        """

        return self._code

    def _set_code(self, flag):
        """
         Sets label to use a code style
         :return: bool
         """

        self._code = flag
        self.style().polish(self)

    def _get_elide_mode(self):
        """
        Returns which elide mode label is using
        :return: Qt.ElideLeft/Qt.ElideMiddle/Qt.ElideRight/Qt.ElideNone
        """

        return self._elide_mode

    def _set_elide_mode(self, value):
        """
        Sets elide mode used by the label
        :param value: Qt.ElideLeft/Qt.ElideMiddle/Qt.ElideRight/Qt.ElideNone
        """

        self._elide_mode = value
        self._update_elided_text()

    theme_type = Property(str, _get_type, _set_type)
    theme_level = Property(int, _get_level, _set_level)
    theme_underline = Property(bool, _get_underline, _set_underline)
    theme_delete = Property(bool, _get_delete, _set_delete)
    theme_mark = Property(bool, _get_mark, _set_mark)
    theme_strong = Property(bool, _get_strong, _set_strong)
    theme_code = Property(bool, _get_code, _set_code)
    theme_elide_mode = Property(bool, _get_elide_mode, _set_elide_mode)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    # def event(self, event):
    #     """
    #     Overrides base QObject event function
    #     """
    #
    #     if event.type() == QEvent.DynamicPropertyChange and event.propertyName() == 'label_text':
    #         self.setText(self.property('label_text'))
    #
    #     return super(BaseLabel, self).event(event)

    def mousePressEvent(self, event):
        self.clicked.emit()
        super(BaseLabel, self).mousePressEvent(event)

    def minimumSizeHint(self):
        """
         Overrides base QObject minimumSizeHint function
         :return: QSize
         """

        return QSize(1, self.fontMetrics().height())

    def resizeEvent(self, event):
        """
        Overrides base QObject resizeEvent function
        """

        self._update_elided_text()

    def text(self):
        """
        Overrides base QLabel text function
        :return: str
        """

        return self._actual_text

    def setText(self, text):
        """
        Overrides base QLabel setText function
        :return: str
        """

        self._actual_text = text
        self._update_elided_text()
        self.setToolTip(text)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def h1(self):
        """
        Sets label with h1 type
        """

        self.theme_level = self.Levels.H1

        return self

    def h2(self):
        """
        Sets label with h2 type
        """

        self.theme_level = self.Levels.H2

        return self

    def h3(self):
        """
        Sets label with h3 type
        """

        self.theme_level = self.Levels.H3

        return self

    def h4(self):
        """
        Sets label with h4 type
        """

        self.theme_level = self.Levels.H4

        return self

    def secondary(self):
        """
        Sets label with secondary type
        """

        self.theme_type = self.Types.SECONDARY

        return self

    def warning(self):
        """
        Sets label with warning type
        """

        self.theme_type = self.Types.WARNING

        return self

    def danger(self):
        """
        Sets label with danger type
        """

        self.theme_type = self.Types.DANGER

        return self

    def strong(self, flag=True):
        """
        Sets label with strong type
        :param flag:
        """

        self.theme_strong = flag

        return self

    def mark(self, flag=True):
        """
        Sets label with mark type
        :param flag:
        """

        self.theme_mark = flag

        return self

    def code(self, flag=True):
        """
        Sets label with code type
        :param flag:
        """

        self.theme_code = flag

        return self

    def delete(self, flag=True):
        """
        Sets label with delete type
        :param flag:
        """

        self.theme_delete = flag

        return self

    def underline(self, flag=True):
        """
        Sets label with underline type
        :param flag:
        """

        self.theme_underline = flag

        return self

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _update_elided_text(self):
        """
        Internal function that updates the elided text on the label
        """

        font_metrics = self.fontMetrics()
        elided_text = font_metrics.elidedText(self._actual_text, self._elide_mode, self.width() - 2 * 2)
        super(BaseLabel, self).setText(elided_text)


class DragDropLine(QLineEdit, object):
    """
    QLineEdit that supports drag and drop behaviour
    """

    def __init__(self, parent=None):
        super(DragDropLine, self).__init__(parent)

        self.setDragEnabled(True)
        self.setReadOnly(True)

    def dragEnterEvent(self, event):
        """
        Overrides QWidget dragEnterEvent to enable drop behaviour with file
        :param event: QDragEnterEvent
        :return:
        """
        data = event.mimeData()
        urls = data.urls()
        if (urls and urls[0].scheme() == 'file'):
            event.acceptProposedAction()

    def dragMoveEvent(self, event):
        data = event.mimeData()
        urls = data.urls()
        if (urls and urls[0].scheme() == 'file'):
            event.acceptProposedAction()

    def dropEvent(self, event):
        data = event.mimeData()
        urls = data.urls()
        if (urls and urls[0].scheme() == 'file'):
            self.setText(urls[0].toLocalFile())


class ClickLabel(QLabel, object):
    """
    This label emits a clicked signal when the user clicks on it
    """

    clicked = Signal()

    def mousePressEvent(self, event):
        self.clicked.emit()
        super(ClickLabel, self).mousePressEvent(event)


class RightElidedLabel(BaseLabel, object):
    """
    Custom label which supports elide right
    """

    def __init__(self, *args, **kwargs):
        super(RightElidedLabel, self).__init__(*args, **kwargs)
        self._text = self.text()

    def setText(self, text):
        """
        Overrides base QLabel setText function
        Used to store original text
        :param text: str
        """

        self._text = text
        super(RightElidedLabel, self).setText(text)

    def resizeEvent(self, event):
        """
        Overrides base QLabel resizeEvent function
        Modifies the text with elided text
        :param event: QResizeEvent
        """

        metrics = QFontMetrics(self.font())
        elided = metrics.elidedText(self._text, Qt.ElideRight, self.width())
        super(RightElidedLabel, self).setText(elided)


class ElidedLabel(BaseLabel, object):
    """
    This label elides text and adds ellipses if the text does not fit
    properly withing the widget frame
    """

    def __init__(self, text='', parent=None):
        super(ElidedLabel, self).__init__(parent=parent)

        self._elide_mode = Qt.ElideRight
        self._actual_text = ""
        self._line_width = 0
        self._ideal_width = None

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        self.setText(text)

    def sizeHint(self):

        base_size_hint = super(ElidedLabel, self).sizeHint()
        return QSize(self._get_width_hint(), base_size_hint.height())

    def _get_width_hint(self):

        if not self._ideal_width:

            doc = QTextDocument()
            try:
                # add the extra space to buffer the width a bit
                doc.setHtml(self._actual_text + "&nbsp;")
                doc.setDefaultFont(self.font())
                width = doc.idealWidth()
            except Exception:
                width = self.width()
            finally:
                qtutils.safe_delete_later(doc)

            self._ideal_width = width

        return self._ideal_width

    def _get_elide_mode(self):
        """
        Returns current elide mode
        """

        return self._elide_mode

    def _set_elide_mode(self, value):
        """
        Set the current elide mode.
        """

        if value != Qt.ElideLeft and value != Qt.ElideRight:
            raise ValueError("elide_mode must be set to either QtCore.Qt.ElideLeft or QtCore.Qt.ElideRight")
        self._elide_mode = value
        self._update_elided_text()

    elide_mode = property(_get_elide_mode, _set_elide_mode)

    def text(self):
        """
        Overridden base method to return the original unmodified text
        """
        return self._actual_text

    def setText(self, text):
        """
        Overridden base method to set the text on the label
        """
        # clear out the ideal width so that the widget can recalculate based on
        # the new text
        self._ideal_width = None
        self._actual_text = text
        self._update_elided_text()

        # if we're elided, make the tooltip show the full text
        if super(ElidedLabel, self).text() != self._actual_text:
            # wrap the actual text in a paragraph so that it wraps nicely
            self.setToolTip("<p>%s</p>" % (self._actual_text,))
        else:
            self.setToolTip("")

    def resizeEvent(self, event):
        """
        Overridden base method called when the widget is resized.
        """

        self._update_elided_text()

    def _update_elided_text(self):
        """
        Update the elided text on the label
        """

        text = self._elide_text(self._actual_text, self._elide_mode)
        QLabel.setText(self, text)

    def _elide_text(self, text, elide_mode):
        """
        Elide the specified text using the specified mode
        :param text:        The text to elide
        :param elide_mode:  The elide mode to use
        :returns:           The elided text.
        """

        # target width is the label width:
        target_width = self.width()

        # Use a QTextDocument to measure html/richtext width
        doc = QTextDocument()
        try:
            doc.setHtml(text)
            doc.setDefaultFont(self.font())

            # if line width is already less than the target width then great!
            line_width = doc.idealWidth()
            if line_width <= target_width:
                self._line_width = line_width
                return text

            # depending on the elide mode, insert ellipses in the correct place
            cursor = QTextCursor(doc)
            ellipses = ""
            if elide_mode != Qt.ElideNone:
                # add the ellipses in the correct place:
                ellipses = "..."
                if elide_mode == Qt.ElideLeft:
                    cursor.setPosition(0)
                elif elide_mode == Qt.ElideRight:
                    char_count = doc.characterCount()
                    cursor.setPosition(char_count - 1)
                cursor.insertText(ellipses)
            ellipses_len = len(ellipses)

            # remove characters until the text fits within the target width:
            while line_width > target_width:

                start_line_width = line_width

                # if string is less than the ellipses length then just return
                # an empty string
                char_count = doc.characterCount()
                if char_count <= ellipses_len:
                    self._line_width = 0
                    return ""

                # calculate the number of characters to remove - should always remove at least 1
                # to be sure the text gets shorter!
                line_width = doc.idealWidth()
                p = target_width / line_width
                # play it safe and remove a couple less than the calculated amount
                chars_to_delete = max(1, char_count - int(float(char_count) * p) - 2)

                # remove the characters:
                if elide_mode == Qt.ElideLeft:
                    start = ellipses_len
                    end = chars_to_delete + ellipses_len
                else:
                    # default is to elide right
                    start = max(0, char_count - chars_to_delete - ellipses_len - 1)
                    end = max(0, char_count - ellipses_len - 1)

                cursor.setPosition(start)
                cursor.setPosition(end, QTextCursor.KeepAnchor)
                cursor.removeSelectedText()

                # update line width:
                line_width = doc.idealWidth()

                if line_width == start_line_width:
                    break

            self._line_width = line_width
            return doc.toHtml()
        finally:
            qtutils.safe_delete_later(doc)

    @property
    def line_width(self):
        """
        :returns: int, width of the line of text in pixels
        """
        return self._line_width


class ThumbnailLabel(QLabel, object):
    def __init__(self, parent=None):
        super(ThumbnailLabel, self).__init__(parent=parent)

    def setPixmap(self, pixmap):
        if pixmap.height() > 55 or pixmap.width() > 80:
            pixmap = pixmap.scaled(QSize(80, 55), Qt.KeepAspectRatio, Qt.SmoothTransformation)
        super(ThumbnailLabel, self).setPixmap(pixmap)


class AnimatedLabel(QLabel, object):
    def __init__(self, text='', duration=800, parent=None):
        super(AnimatedLabel, self).__init__(text, parent=parent)

        font = self.font()
        font.setPointSize(18)
        font.setBold(True)
        self.setFont(font)

        self.opacity_effect = graphicseffects.OpacityEffect(target=self, duration=duration)

        self.opacity_effect.animation.valueChanged.connect(self._set_label_opacity)
        self.opacity_effect.animation.finished.connect(self.hide)

    # region Properties
    def get_duration(self):
        return self.opacity_effect.duration

    def set_duration(self, value):
        self.opacity_effect.duration = value

    duration = property(get_duration, set_duration)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setOpacity(self.opacity_effect.animation.currentValue())
        if self.text():
            painter.drawText(event.rect(), self.text())

    def show_message(self, text):
        self.setText(text)
        parent = self.parent()
        if parent:
            # TODO: Add option to set aligment of the text inside its parent
            self.move(10, parent.size().height() - self.fontMetrics().height() - 10)
            # if isinstance(parent, QAbstractScrollArea) and parent.verticalScrollBar().isVisible():
            #     self.move(
            #     parent.size().width() - self.fontMetrics().width(text) - 10 - parent.verticalScrollBar().width(), 0)
            # else:
            #     self.move(parent.size().width() - self.fontMetrics().width(text) - 10, 0)
        self.opacity_effect.fade_in_out()

    def _set_label_opacity(self, value):
        if self.opacity_effect.animation.state() == QPropertyAnimation.Running:
            self.setVisible(value > 0)
            self.update()


class IconLabel(QLabel, object):
    """
    Displays an icon beside a standard label
    """

    clicked = Signal()

    def __init__(self, *args, **kwargs):
        self._icon = kwargs.pop('icon') if 'icon' in kwargs else None
        self._press = False
        self._two_sided = True

        super(IconLabel, self).__init__(*args, **kwargs)

        self.setContentsMargins(20, 0, 0, 0)
        self._value_a = self.text()
        self._value_b = self.text()

    def get_value_a(self):
        return self._value_a

    def get_value_b(self):
        return self._value_b

    value_a = property(get_value_a)
    value_b = property(get_value_b)

    def mousePressEvent(self, event):
        if self._two_sided:
            self._press = True
            self.setText(self._value_a, False)
            self.update()
        self.clicked.emit()

    def mouseReleaseEvent(self, event):
        if self._two_sided:
            self._press = False
            self.setText(self._value_b, False)
            self.update()

    def paintEvent(self, event):
        p = QPainter(self)
        if self._icon:
            self._icon.paint(p, QRect(0, self.height() * 0.5 - 5, 12, 12))
        super(IconLabel, self).paintEvent(event)

    def setText(self, text, update=True):
        super(IconLabel, self).setText('<font>%s</font>' % text)
        if update:
            self._value_b = text

    def set_icon(self, icon):
        self._icon = icon


class ClippedLabel(QLabel, object):
    def __init__(self, text='', width=0, ellipsis=True, always_show_all=False, parent=None):
        super(ClippedLabel, self).__init__(text, parent)

        self._always_show_all = always_show_all
        self._ellipsis = ellipsis

        self._width = None
        self._text = None
        self._elided = None

        self.setMinimumWidth(width if width else 0)

    def paintEvent(self, event):
        painter = QPainter(self)
        self.drawFrame(painter)
        margin = self.margin()
        rect = self.contentsRect()
        rect.adjust(margin, margin, -margin, -margin)
        text = self.text()
        width = rect.width()
        if text != self._text or width != self._width:
            self._text = text
            self._width = width
            self._elided = self.fontMetrics().elidedText(text, Qt.ElideRight, width)
        option = QStyleOption()
        option.initFrom(self)

        if self._always_show_all:
            if self._width >= self.sizeHint().width():
                self.style().drawItemText(
                    painter, rect, self.alignment(), option.palette, self.isEnabled(),
                    self.text(), self.foregroundRole())
        else:
            if self._ellipsis:
                self.style().drawItemText(
                    painter, rect, self.alignment(), option.palette, self.isEnabled(),
                    self._elided, self.foregroundRole())
            else:
                self.style().drawItemText(
                    painter, rect, self.alignment(), option.palette, self.isEnabled(),
                    self.text(), self.foregroundRole())



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different group widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Signal, Property
from Qt.QtWidgets import QSizePolicy, QWidget, QFrame, QBoxLayout, QGroupBox, QButtonGroup
from Qt.QtGui import QIcon

from tpDcc.managers import resources
from tpDcc.libs.python import decorators, python
from tpDcc.libs.resources.core import theme
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import layouts, buttons


class BaseGroup(QGroupBox, object):
    def __init__(self, name='', parent=None, layout_spacing=2, layout_orientation=Qt.Vertical):

        self._layout_orientation = layout_orientation

        super(BaseGroup, self).__init__(parent)

        self.setTitle(name)

        if layout_orientation == Qt.Horizontal:
            self.main_layout = layouts.HorizontalLayout(spacing=layout_spacing, margins=(2, 2, 2, 2))
        else:
            self.main_layout = layouts.VerticalLayout(spacing=layout_spacing, margins=(2, 2, 2, 2))
        self.main_layout.setAlignment(Qt.AlignCenter)
        self.setLayout(self.main_layout)

        self.ui()
        self.setup_signals()

    def ui(self):
        """
        Function that sets up the ui of the widget
        Override it on new widgets
        """

        pass

    def setup_signals(self):
        """
        Function that set up signals of the group widgets
        """

        pass

    def set_title(self, new_title):
        """
        Set the title of the group
        """

        self.setTitle(new_title)

    def addWidget(self, widget):
        """
        Adds a a new widget to the group box layout
        NOTE: We do not follow the nomenclature to make the call similar to Qt
        :param widget: QWidget
        """

        if not widget:
            return

        self.main_layout.addWidget(widget)

    def addLayout(self, layout):
        """
        Adds a a new layout to the group box layout
        NOTE: We do not follow the nomenclature to make the call similar to Qt
        :param widget: QLayout
        """

        if not layout:
            return

        self.main_layout.addLayout(layout)


class CollapsableGroup(BaseGroup, object):
    def __init__(self, name='', parent=None, collapsable=True):
        self._collapsable = collapsable
        super(CollapsableGroup, self).__init__(name, parent, layout_orientation=Qt.Horizontal)

    def ui(self):
        super(CollapsableGroup, self).ui()

        self._base_widget = QWidget()
        if self._layout_orientation == Qt.Vertical:
            manager_layout = layouts.VerticalLayout(spacing=2, margins=(4, 4, 4, 4))
        else:
            manager_layout = layouts.HorizontalLayout(spacing=2, margins=(4, 4, 4, 4))
        manager_layout.setAlignment(Qt.AlignCenter)
        self._base_widget.setLayout(manager_layout)
        self.main_layout.addWidget(self._base_widget)
        self.main_layout = manager_layout

    def mousePressEvent(self, event):
        super(CollapsableGroup, self).mousePressEvent(event)

        if not event.button() == Qt.LeftButton:
            return

        if self._collapsable:
            if event.y() < 30:
                if self._base_widget.isHidden():
                    self.expand_group()
                else:
                    self.collapse_group()

    def set_collapsable(self, flag):
        """
        Sets if the group can be collapsed or not
        :param flag: bool
        """

        self._collapsable = flag

    def set_title(self, title):
        if not title.startswith('+ '):
            title = '+ ' + title
        self.setTitle(title)

    def expand_group(self):
        """
        Expands the content of the group
        """

        self.setVisible(True)
        title = self.title()
        title = title.replace('+', '-')
        self.setTitle(title)

    def collapse_group(self):
        """
        Collapse the content of the group
        """

        self._base_widget.setVisible(False)
        title = self.title()
        title = title.replace('-', '+')
        self.setTitle(title)


@theme.mixin
class GroupBoxWidget(base.BaseFrame):

    toggled = Signal(bool)

    def __init__(self, title, widget=None, persistent=False, settings=None, *args, **kwargs):

        self._title = title
        self._widget = None
        self._persistent = None
        self._settings = settings

        super(GroupBoxWidget, self).__init__(*args, **kwargs)

        if widget:
            self.set_widget(widget)
            # We force the update of the check status to make sure that the wrapped widget visibility is updated
            self.set_checked(self.is_checked())

        self.set_persistent(persistent)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def get_main_layout(self):
        return layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))

    def ui(self):
        super(GroupBoxWidget, self).ui()

        self._title_widget = buttons.BaseButton(self._title, parent=self)
        self._title_widget.setCheckable(True)

        self._on_icon = resources.icon('down_button')
        self._off_icon = resources.icon('right_button')
        self._title_widget.setIcon(self._off_icon)

        self._widget_frame = QFrame(self)
        self._widget_frame.setObjectName('contentsWidget')
        widget_frame_layout = layouts.VerticalLayout(spacing=2, margins=(0, 0, 0, 0))
        self._widget_frame.setLayout(widget_frame_layout)

        self.main_layout.addWidget(self._title_widget)
        self.main_layout.addWidget(self._widget_frame)

    def setup_signals(self):
        self._title_widget.toggled.connect(self._on_toggled_title)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def is_checked(self):
        """
        Returns whether or not group box is checked
        :return: bool
        """

        return self._title_widget.isChecked()

    def set_checked(self, flag):
        """
        Sets the check statue of the group box
        :param flag: bool
        """

        self._title_widget.setChecked(flag)
        self._title_widget.setIcon(self._on_icon if flag else self._off_icon)
        self._widget_frame.setVisible(flag)
        if self._widget:
            self._widget.setVisible(flag)

    def is_persistent(self):
        """
        Returns whether or not widget state is stored in settings
        :return: bool
        """

        return self._persistent

    def set_persistent(self, flag):
        """
        Sets whether or not widget state is stored in settings
        :param flag: bool
        """

        self._persistent = flag
        self.load_settings()

    def title(self):
        """
        Returns group box title
        :return: str
        """

        return self._title_widget.text()

    def set_widget(self, widget):
        """
        Sets the widget to hide when the user clicks the title
        :param widget: QWidget
        """

        self._widget = widget
        self._widget.setParent(self._widget_frame)
        self._widget_frame.layout().addWidget(self._widget)

    # ============================================================================================================
    # SETTINGS
    # ============================================================================================================

    def load_settings(self):
        """
        Loads widget state from given settings
        """

        if not self._settings or not self._persistent:
            return
        if not self.objectName():
            raise NameError('Impossible to save "{}" widget state because no objectName is defined!'.format(self))

        data = {self.objectName(): {'checked': self.is_checked()}}
        self._settings.save(data)

    def save_settings(self):
        """
        Saves current widget state into settings
        """

        if not self._settings or not self._persistent:
            return
        if not self.objectName():
            raise NameError('Impossible to load "{}" widget state because no objectName is defined!'.format(self))

        data = self._settings.read()
        data = data.get(self.objectName(), dict())
        if data and isinstance(data, dict):
            checked = data.get('checked', True)
            self.set_checked(checked)

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_toggled_title(self, flag):
        """
        Internal callback function that is called each time title group widget is toggled
        :param flag: bool
        """

        self.save_settings()
        self.set_checked(flag)
        self.toggled.emit(flag)


class BaseButtonGroup(base.BaseWidget, object):
    def __init__(self, orientation=Qt.Horizontal, parent=None):

        self._orientation = 'horizontal' if orientation == Qt.Horizontal else 'vertical'

        super(BaseButtonGroup, self).__init__(parent=parent)

    def get_main_layout(self):
        main_layout = QBoxLayout(
            QBoxLayout.LeftToRight if self._orientation == 'horizontal' else QBoxLayout.TopToBottom)
        main_layout.setContentsMargins(0, 0, 0, 0)

        return main_layout

    def ui(self):
        super(BaseButtonGroup, self).ui()

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self._button_group = QButtonGroup()

    @decorators.abstractmethod
    def create_button(self, data_dict):
        """
        Must be implemented in custom button groups
        Creates a new button for this group
        :param data_dict: dict
        :return: new button instance
        """

        raise NotImplementedError(
            'Function create_button for class "{}" is not implemented!'.format(self.__class__.__name__))

    def get_button_group(self):
        """
        Returns button group internal object
        :return: QButtonGroup
        """

        return self._button_group

    def clear(self):
        """
        Clears all buttons contained in this group
        """

        for btn in self._button_group.buttons():
            self._button_group.removeButton(btn)
            self.main_layout.removeWidget(btn)
            btn.setVisible(False)
            btn.deleteLater()

    def add_button(self, data_dict, index=None):
        """
        Adds a new button to this group
        :param data_dict: dict
        :param index: int or None
        :return: new added button
        """

        if python.is_string(data_dict):
            data_dict = {'text': data_dict}
        elif isinstance(data_dict, QIcon):
            data_dict = {'icon': data_dict}

        new_btn = self.create_button(data_dict)
        new_btn.setProperty('combine', self._orientation)

        if data_dict.get('text'):
            new_btn.setProperty('text', data_dict.get('text'))
        if data_dict.get('icon'):
            new_btn.setProperty('icon', data_dict.get('icon'))
        if data_dict.get('data'):
            new_btn.setProperty('data', data_dict.get('data'))
        if data_dict.get('checked'):
            new_btn.setProperty('checked', data_dict.get('checked'))
        if data_dict.get('shortcut'):
            new_btn.setProperty('shortcut', data_dict.get('shortcut'))
        if data_dict.get('tooltip'):
            new_btn.setProperty('toolTip', data_dict.get('tooltip'))
        if data_dict.get('clicked'):
            new_btn.clicked.connect(data_dict.get('clicked'))
        if data_dict.get('toggled'):
            new_btn.toggled.connect(data_dict.get('toggled'))

        if index is None:
            self._button_group.addButton(new_btn)
        else:
            self._button_group.addButton(new_btn, index)

        if self.main_layout.count() == 0:
            new_btn.setChecked(True)

        self.main_layout.insertWidget(self.main_layout.count(), new_btn)

        return new_btn

    def set_button_list(self, button_list):
        """
        Empties group and add all buttons given in the list of buttons
        :param button_list: list(dict)
        """

        self.clear()

        for index, data_dict in enumerate(button_list):
            new_btn = self.add_button(data_dict=data_dict, index=index)
            if index == 0:
                new_btn.setProperty('position', 'left')
            elif index == len(button_list) - 1:
                new_btn.setProperty('position', 'right')
            else:
                new_btn.setProperty('position', 'center')


class PushButtonGroup(BaseButtonGroup, object):
    def __init__(self, orientation=Qt.Horizontal, parent=None):
        super(PushButtonGroup, self).__init__(orientation=orientation, parent=parent)

        self._type = buttons.BaseButton.Types.PRIMARY
        self._size = theme.Theme.DEFAULT_SIZE
        self._button_group.setExclusive(True)
        self.set_spacing(1)

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = value

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value):
        self._size = value

    def create_button(self, data_dict):
        """
        Implements BaseButtonGroup create_button abstract function
        :param data_dict:
        :return:
        """

        new_btn = buttons.StyleBaseButton()
        new_btn.size = data_dict.get('size', self._size)
        new_btn.type = data_dict.get('type', self._type)

        return new_btn


class RadioButtonGroup(BaseButtonGroup, object):
    checkedChanged = Signal(int)

    def __init__(self, orientation=Qt.Horizontal, parent=None):
        super(RadioButtonGroup, self).__init__(orientation=orientation, parent=parent)

        self._button_group.setExclusive(True)
        self.set_spacing(15)

    def setup_signals(self):
        self._button_group.buttonClicked.connect(self.checkedChanged)

    def create_button(self, data_dict):
        """
        Implements BaseButtonGroup create_button abstract function
        :param data_dict:
        :return:
        """

        return buttons.BaseRadioButton()

    def _get_checked(self):
        return self._button_group.checkedId()

    def _sert_checked(self, value):
        btn = self._button_group.button(value)
        if btn:
            btn.setChecked(True)
            self.checkedChanged.emiet(value)

    checked = Property(int, _get_checked, _sert_checked, notify=checkedChanged)
